# Copyright (C) 2016 R. Warmbier Materials for Energy Research Group,
# Wits University

import copy
import numpy as np
from ase.dft.kpoints import monkhorst_pack, get_monkhorst_pack_size_and_offset
from gpaw import KPointError
from gpaw.kpt_descriptor import KPointDescriptor

"""
This file provides routines to create non-uniform k-point grids. We use
locally uniform grids of different densities, practically grid refinement.

One starts from a standard Monkhorst-Pack grid and chooses a number of grid
points, which shall be replaced with a finer grid of a given size.

The user can further choose, whether the original symmetry of the system is
enforced (careful!) or whether a reduced symmetry should be used.

Optionally, the user can ask to add all k+q points, if not already included.

Please cite https://doi.org/10.1016/j.cpc.2018.03.001

Example (Graphene):

calc = GPAW(mode=PW(ecut=450),
            kpts={"size":[15,15,1], "gamma":True},
            kpt_refine={"center":[1./3,1./3,0.], "size":[3,3,1],
                        "reduce_symmetry":False,
                        "q":[1./15,1./15,0.]},
           )

"""


def create_kpoint_descriptor_with_refinement(refine, bzkpts_kc, nspins, atoms,
                                             symmetry, comm, timer):
    """Main routine to build refined k-point grids."""
    if 'center' not in refine:
        raise RuntimeError('Center for refinement not given!')
    if 'size' not in refine:
        raise RuntimeError('Grid size for refinement not given!')

    center_ic = np.array(refine.get('center'), dtype=float, ndmin=2)
    size = np.array(refine.get('size'), ndmin=2)
    reduce_symmetry = refine.get('reduce_symmetry', True)

    # Check that all sizes are odd. That's not so much an issue really.
    # But even Monkhorst-Pack grids have points on the boundary,
    # which just would require more special casing, which I want to avoid.
    if (np.array(size) % 2 == 0).any():
        raise RuntimeError('Grid size for refinement must be odd!  Is: {}'.
                           format(size))

    # Arguments needed for k-point descriptor construction
    kwargs = {'nspins': nspins, 'atoms': atoms, 'symmetry': symmetry,
              'comm': comm}

    # Define coarse grid points
    bzk_coarse_kc = bzkpts_kc

    # Define fine grid points
    centers_i, bzk_fine_kc, weight_fine_k = get_fine_bzkpts(center_ic, size,
                                                            bzk_coarse_kc,
                                                            kwargs)

    if reduce_symmetry:
        # Define new symmetry object ignoring symmetries violated by the
        # refined kpoints
        kd_fine = create_kpoint_descriptor(bzk_fine_kc, **kwargs)
        symm = prune_symmetries_kpoints(kd_fine, symmetry)
        del kd_fine
    else:
        symm = copy.copy(symmetry)
    kwargs['symmetry'] = symm

    # Create new descriptor with both sets of points

    with timer('Create mixed descriptor'):
        kd = create_mixed_kpoint_descriptor(bzk_coarse_kc,
                                            bzk_fine_kc,
                                            centers_i,
                                            weight_fine_k,
                                            kwargs)

    # Add missing k-points to fulfill group properties with
    # zero-weighted points
    with timer('Add_missing_points'):
        kd = add_missing_points(kd, kwargs)

    # Add additional +q k-points, if necessary
    if 'q' in refine:
        timer.start("+q")
        N_coarse_c = get_monkhorst_pack_size_and_offset(bzk_coarse_kc)[0]
        bla = N_coarse_c * refine['q']
        if not max(abs(bla - np.rint(bla))) < 1e-8:
            kd.refine_info.almostoptical = True
        kd = add_plusq_points(kd, refine['q'], kwargs)
        symm = kd.symmetry
        kwargs['symmetry'] = symm
        kd = add_missing_points(kd, kwargs)
        timer.stop("+q")

    return kd


def create_kpoint_descriptor(bzkpts_kc, nspins, atoms, symmetry, comm):
    kd = KPointDescriptor(bzkpts_kc, nspins)
    # self.timer.start('Set symmetry')
    kd.set_symmetry(atoms, symmetry, comm=comm)
    # self.timer.stop('Set symmetry')
    return kd


def create_mixed_kpoint_descriptor(bzk_coarse_kc, bzk_fine_kc, centers_i,
                                   weight_fine_k, kwargs):
    assert len(weight_fine_k) == bzk_fine_kc.shape[0]

    # Get old Monkhorst-Pack grid points and delete the centers from them
    nbzkpts_coarse = bzk_coarse_kc.shape[0]
    bzk_new_kc = np.delete(bzk_coarse_kc, centers_i, axis=0)
    nbzkpts_coarse_new = bzk_new_kc.shape[0]
    # Add refined points
    nbzkpts_fine = bzk_fine_kc.shape[0]
    bzk_new_kc = np.append(bzk_new_kc, bzk_fine_kc, axis=0)

    # Construct the new KPointDescriptor
    kd_new = create_kpoint_descriptor(bzk_new_kc, **kwargs)
    refine_info = KRefinement()
    refine_info.set_unrefined_nbzkpts(nbzkpts_coarse)
    label_k = np.array(nbzkpts_coarse_new * ['mh'] + nbzkpts_fine * ['refine'])
    refine_info.set_label_k(label_k)
    weight_k = np.append(np.ones(nbzkpts_coarse_new), weight_fine_k)
    refine_info.set_weight_k(weight_k)
    kd_new.refine_info = refine_info
    assert len(weight_k) == bzk_new_kc.shape[0]

    # This new Descriptor is good, except the weights are completely wrong now.
    kd_new.weight_k = np.bincount(kd_new.bz2ibz_k, weight_k) / nbzkpts_coarse
    assert np.abs(np.sum(kd_new.weight_k) - 1) < 1e-6

    # return kd_new.copy()
    return kd_new


def get_fine_bzkpts(center_ic, size, bzk_coarse_kc, kwargs):
    """Return list of reducible refined kpoints and the indexes for the points
    on the coarse kpoint grid. """

    # Define Monkhorst-Pack grid as starting point
    kd_coarse = create_kpoint_descriptor(bzk_coarse_kc, **kwargs)

    # Determine how many (cubic) nearest neighbour shells to replace
    nshells = size.shape[0]

    # Get coordinates of neighbour k-points. First index is the 'shell' aka
    # order of neighbour. Coordinates always centred around zero.
    neighbours_kc = construct_neighbours_by_shells(nshells, kd_coarse.N_c)

    # loop over all the different shells
    centers_i = np.empty((0), dtype=int)
    bzk_fine_kc = []
    weight_k = []
    for shell in range(nshells):
        # Determine all k-points in shell, which will be refinement centers
        shell_kc = []
        for i, center_c in enumerate(center_ic):
            shell_c = neighbours_kc[shell] + center_c
            shell_kc.append(shell_c)
        shell_kc = np.concatenate(shell_kc)

        # Find all equivalent centers using full symmetry
        center_i = []
        for k, shell_c in enumerate(shell_kc):
            equiv_i = find_equivalent_kpoints(shell_c, kd_coarse)
            center_i.append(equiv_i)
        center_i = np.concatenate(center_i)

        # Remove redundant entries, which come from symmetry
        center_i, index_map = np.unique(center_i, return_index=True)

        # Append to list for all shells
        # Remove also points which occur also in a previous shell. This can
        # happen due to symmetry or overlap of centers' shells
        center_i = np.setdiff1d(center_i, centers_i)
        centers_i = np.append(centers_i, center_i)

        # Assign the kpoint vectors corresponding to the indexes
        centers_kc = bzk_coarse_kc[center_i]

        # Get scaled Monkhorst-Pack for refinement
        mh_kc = get_reduced_monkhorst(size[shell], kd_coarse.N_c)

        # Gather the refined points of the grid for all centers
        for k, centers_c in enumerate(centers_kc):
            this_kc = mh_kc + centers_c
            bzk_fine_kc.append(this_kc)

        # Determine weight of refined points
        weight = 1. / np.prod(size[shell])
        nkpts = len(center_i) * np.prod(size[shell])
        weight *= np.ones(nkpts)
        weight_k.append(weight)

    weight_k = np.concatenate(weight_k)
    bzk_fine_kc = np.concatenate(bzk_fine_kc)

    assert np.abs(len(centers_i) - sum(weight_k)) < 1e-6
    assert len(weight_k) == bzk_fine_kc.shape[0]

    return centers_i, bzk_fine_kc, weight_k


def find_equivalent_kpoints(point, kd):
    """Find coordinates and BZ index for all refinement centers."""

    # check whether point in bzkpts_kc
    min_dist, k = minimal_point_distance(point, kd.bzk_kc)
    if min_dist > 1e-8:
        raise RuntimeError('Could not find k-point in kd list!')

    equiv_k = list(set(kd.bz2bz_ks[k]))
    # equiv_kc = kd.bzk_kc[equiv_k]

    return np.array(equiv_k)  # , equiv_kc


def get_reduced_monkhorst(size, N_c):
    """Find monkhorst_pack according to size of refined grid and shrink it into
    the volume of one original kpoint."""

    mh_kc = monkhorst_pack(size)
    return mh_kc / N_c


def construct_neighbours_by_shells(nshells, N_c):
    """Construct a list of neighbours (translations from center for each
    shell around the center."""

    neighbours = []

    # The innermost shell is always the center itself
    neighbours.append(np.array([0, 0, 0], dtype=float, ndmin=2))

    # Loop through the other shells
    for shell in range(1, nshells):
        # Construct the displacements/elements for each dimension.
        elements = []
        for i in range(3):
            if N_c[i] == 1:
                elements.append([0, ])
            else:
                elements.append(list(range(-shell, shell + 1)))

        # Construct the vectors
        # For each valid point, at least one component must have the value
        # of the shell index (+ or -).
        this_list = []
        for cx in elements[0]:
            for cy in elements[1]:
                for cz in elements[2]:
                    candidate = [cx, cy, cz]
                    if (np.abs(np.array(candidate)) == shell).any():
                        this_list.append(candidate)

        this_list = np.array(this_list, dtype=float, ndmin=2) / N_c
        neighbours.append(this_list)

    return np.array(neighbours)


def prune_symmetries_kpoints(kd, symmetry):
    """Remove the symmetries which are violated by a kpoint set."""

    new_symmetry = copy.copy(symmetry)

    nsyms = len(symmetry.op_scc)

    where = np.where(~np.any(kd.bz2bz_ks[:, 0:nsyms] == -1, 0))

    new_symmetry.op_scc = new_symmetry.op_scc[where]
    new_symmetry.ft_sc = new_symmetry.ft_sc[where]
    new_symmetry.a_sa = new_symmetry.a_sa[where]

    inv_cc = -np.eye(3, dtype=int)
    new_symmetry.has_inversion = (new_symmetry.op_scc ==
                                  inv_cc).all(2).all(1).any()

    return new_symmetry


def find_missing_points(kd):
    """Find points in k-point descriptor, which miss in the set to fill the
    group."""
    if -1 not in kd.bz2bz_ks:
        return None

    # Find unaccounted points
    buf = np.empty((0, 3))
    for k in range(kd.bz2bz_ks.shape[0]):
        for s in range(kd.bz2bz_ks.shape[1]):
            if kd.bz2bz_ks[k, s] == -1:
                sign = 1.0
                i = s
                if kd.symmetry.time_reversal:
                    if s / kd.bz2bz_ks.shape[1] >= 0.5:
                        i = s - int(kd.bz2bz_ks.shape[1] / 2)
                        sign = -1.
                k_c = sign * np.dot(kd.symmetry.op_scc[i], kd.bzk_kc[k])
                k_c -= np.rint(k_c)
                # Check that k_c hasn't been added yet
                if buf.shape[0] > 0:
                    min_dist, index = minimal_point_distance(k_c, buf)
                    if min_dist < 1e-6:
                        continue
                buf = np.concatenate((buf, np.array(k_c, ndmin=2)))

    return buf


def add_missing_points(kd, kwargs):
    """Add points to k-point descriptor, which miss to fill the group."""
    add_points_kc = find_missing_points(kd)
    if add_points_kc is None:
        return kd

    kd_new = create_new_descriptor_with_zero_points(kd, add_points_kc, kwargs)
    assert -1 not in kd_new.bz2bz_ks

    return kd_new


def add_plusq_points(kd, q_c, kwargs):
    """Add +q points to k-point descriptor, if missing. Also, reduce the
    symmetry of the system as necessary."""

    # Add missing points to retrieve full symmetry. Might be redundant.
    _kd = add_missing_points(kd, kwargs)

    # Find missing q
    add_points_kc = []
    for k in range(_kd.nbzkpts):
        # if q_c is small, use q_c = 0.0 for mh points, else they don't need
        # extra points anyway
        if _kd.refine_info.label_k[k] == 'mh':
            continue
        try:
            _kd.find_k_plus_q(q_c, [k])
        except KPointError:
            k_c = _kd.bzk_kc[k] + q_c
            add_points_kc.append(np.array(k_c, ndmin=2))
    add_points_kc = np.concatenate(add_points_kc)

    if add_points_kc.shape[0] == 0:
        return _kd

    # Find reduced +q symmetry
    bzk_kc = np.append(_kd.bzk_kc, add_points_kc, axis=0)
    _kd_new = create_kpoint_descriptor(bzk_kc, **kwargs)
    symm_new = prune_symmetries_kpoints(_kd_new, kwargs.get('symmetry'))
    kwargs['symmetry'] = symm_new
    del _kd_new, _kd

    kd_new = create_new_descriptor_with_zero_points(kd, add_points_kc, kwargs)
    return kd_new


def create_new_descriptor_with_zero_points(kd, add_points_kc, kwargs):
    """Create a new k-point descriptor including some additional zero-weighted
    points."""
    bzk_kc = np.append(kd.bzk_kc, add_points_kc, axis=0)
    nbzkpts_add = add_points_kc.shape[0]

    # Make sure all points are unique
    assert bzk_kc.shape[0] == np.vstack([tuple(row)
                                         for row in bzk_kc]).shape[0]

    # Construct the new KPointDescriptor
    kd_new = create_kpoint_descriptor(bzk_kc, **kwargs)

    # Update refine_info
    kd_new.refine_info = kd.refine_info.copy()
    label_k = np.append(kd.refine_info.label_k,
                        np.array(nbzkpts_add * ['zero']))
    kd_new.refine_info.set_label_k(label_k)
    # Avoid exact zero, as that would screw up the occupations calculation
    weight_k = np.append(kd.refine_info.weight_k, 1e-10 * np.ones(nbzkpts_add))
    kd_new.refine_info.set_weight_k(weight_k)

    # Correct ibz weights
    kd_new.weight_k = np.bincount(kd_new.bz2ibz_k, weight_k)
    kd_new.weight_k *= 1.0 / kd_new.refine_info.get_unrefined_nbzkpts()
    assert np.abs(np.sum(kd_new.weight_k) - 1) < 1e-6

    return kd_new


def minimal_point_distance(point_c, bzk_kc):
    d_kc = point_c - bzk_kc
    d_k = abs(d_kc - d_kc.round()).sum(1)
    k = d_k.argmin()
    return d_k[k], k


class KRefinement:
    """Additional information for refined kpoint grids.
    """
    def __init__(self):
        self.mhnbzkpts = None
        self.label_k = None
        self.weight_k = None
        self.almostoptical = None

    def __str__(self):
        s = "Using k-point grid refinement"
        if self.almostoptical:
            s += " with almostoptical approximation"
        s += "\n"
        return s

    def set_unrefined_nbzkpts(self, mhnbzkpts):
        self.mhnbzkpts = mhnbzkpts

    def get_unrefined_nbzkpts(self):
        return self.mhnbzkpts

    def set_weight_k(self, weight_k):
        self.weight_k = weight_k

    def get_weight_k(self):
        return self.weight_k

    def set_label_k(self, label_k):
        self.label_k = label_k

    def get_label_k(self):
        return self.label_k

    def copy(self):
        refine_info = KRefinement()
        refine_info.mhnbzkpts = self.mhnbzkpts
        refine_info.label_k = self.label_k
        refine_info.weight_k = self.weight_k
        refine_info.almostoptical = self.almostoptical
        return refine_info



import numpy as np
import pytest
from functools import partial

from ase.units import Ha
from ase.data import chemical_symbols

from gpaw.response import ResponseContext, ResponseGroundStateAdapter
from gpaw.response.pair_functions import SingleQPWDescriptor
from gpaw.response.paw import (calculate_pair_density_correction,
                               calculate_matrix_element_correction)
from gpaw.response.site_paw import calculate_site_matrix_element_correction
from gpaw.response.localft import extract_micro_setup, add_LSDA_trans_fxc

from gpaw.setup import create_setup
from gpaw.sphere.rshe import calculate_reduced_rshe


def pawdata():
    for symbol in chemical_symbols:
        try:
            setup = create_setup(symbol)
        except FileNotFoundError:
            pass
        else:
            yield setup


@pytest.mark.response
@pytest.mark.serial
@pytest.mark.parametrize('pawdata', pawdata())
def test_paw_corrections(pawdata):
    G_Gv = np.zeros((5, 3))
    G_Gv[:, 0] = np.linspace(0, 20, 5)
    calculate_pair_density_correction(G_Gv, pawdata=pawdata)


@pytest.mark.response
@pytest.mark.serial
def test_paw_correction_consistency(gpw_files):
    """Test consistency of the pair density PAW corrections."""
    context = ResponseContext()
    gs = ResponseGroundStateAdapter.from_gpw_file(gpw_files['fe_pw'],
                                                  context=context)

    # Set up plane-wave basis
    ecut = 50  # eV
    qpd = SingleQPWDescriptor.from_q([0., 0., 0.5],  # N-point
                                     ecut / Ha, gs.gd)
    qG_Gv = qpd.get_reciprocal_vectors(add_q=True)

    # Calculate ordinary pair density corrections
    pawdata = gs.pawdatasets[0]
    Q1_Gii = calculate_pair_density_correction(qG_Gv, pawdata=pawdata)

    # Calculate pair density as a generalized matrix element
    # Expand unity in real spherical harmonics
    rgd = pawdata.xc_correction.rgd
    Y_nL = pawdata.xc_correction.Y_nL
    f_ng = np.ones((Y_nL.shape[0], rgd.N))
    rshe, _ = calculate_reduced_rshe(rgd, f_ng, Y_nL, lmax=0)
    # Calculate correction
    Q2_Gii = calculate_matrix_element_correction(qG_Gv, pawdata, rshe)

    assert Q2_Gii == pytest.approx(Q1_Gii, rel=1e-3, abs=1e-5)


@pytest.mark.response
@pytest.mark.serial
def test_site_paw_correction_consistency(gpw_files):
    """Test consistency of generalized matrix elements."""
    context = ResponseContext()
    gs = ResponseGroundStateAdapter.from_gpw_file(gpw_files['fe_pw'],
                                                  context=context)

    # Calculate and expand the LDA fxc kernel in real spherical harmonics
    pawdata = gs.pawdatasets[0]
    micro_setup = extract_micro_setup(gs, 0)
    add_fxc = partial(add_LSDA_trans_fxc, fxc='ALDA')
    fxc_ng = micro_setup.evaluate_function(add_fxc)
    rshe, _ = calculate_reduced_rshe(micro_setup.rgd, fxc_ng,
                                     micro_setup.Y_nL, wmin=1e-8)

    # Calculate PAW correction with G + q = 0
    qG_Gv = np.zeros((1, 3))
    nF_Gii = calculate_matrix_element_correction(qG_Gv, pawdata, rshe)

    # Calculate PAW correction with site cutoff exceeding the augmentation
    # sphere radius
    augr = gs.get_aug_radii()[0]
    rcut_p = [augr * 1.5]
    drcut = augr * 0.25
    lambd_p = [0.5]
    nF_pii = calculate_site_matrix_element_correction(pawdata, rshe,
                                                      rcut_p, drcut, lambd_p)

    assert nF_Gii[0] == pytest.approx(nF_pii[0])



from math import sqrt
import pytest
from ase import Atoms
from gpaw import GPAW, Mixer, Davidson
from gpaw.test import equal
from gpaw.xc.vdw import VDWFunctional


@pytest.mark.libxc
def test_vdw_ar2(in_tmp_dir):
    energy_tolerance = 0.002

    def test():
        vdw = VDWFunctional('vdW-DF', verbose=1)
        d = 3.9
        x = d / sqrt(3)
        L = 3.0 + 2 * 4.0
        dimer = Atoms('Ar2', [(0, 0, 0), (x, x, x)], cell=(L, L, L))
        dimer.center()
        calc = GPAW(mode='fd', h=0.2, xc=dict(name='revPBE', stencil=1),
                    mixer=Mixer(0.8, 7, 50.0),
                    eigensolver=Davidson(5))
        dimer.calc = calc
        e2 = dimer.get_potential_energy()
        calc.write('Ar2.gpw')
        e2vdw = calc.get_xc_difference(vdw)
        e2vdwb = GPAW('Ar2.gpw').get_xc_difference(vdw)
        print(e2vdwb - e2vdw)
        assert abs(e2vdwb - e2vdw) < 1e-9
        del dimer[1]
        e = dimer.get_potential_energy()
        evdw = calc.get_xc_difference(vdw)

        E = 2 * e - e2
        Evdw = E + 2 * evdw - e2vdw
        print(E, Evdw)
        assert abs(E - -0.0048) < 6e-4, abs(E)
        assert abs(Evdw - +0.0223) < 3e-3, abs(Evdw)

        print(e2, e)
        equal(e2, -0.005, energy_tolerance)
        equal(e, -0.005, energy_tolerance)

    test()



"""Test GW band-gaps for Si."""

import pytest
from ase.build import bulk

from gpaw import GPAW, PW, FermiDirac
from gpaw.mpi import world
from gpaw.response.g0w0 import G0W0


def generate_si_systems():
    a = 5.43
    si1 = bulk('Si', 'diamond', a=a)
    si2 = si1.copy()
    si2.positions -= a / 8

    return [si1, si2]


def run(atoms, symm, nblocks):
    atoms.calc = GPAW(mode=PW(250),
                      eigensolver='rmm-diis',
                      occupations=FermiDirac(0.01),
                      symmetry=symm,
                      kpts={'size': (2, 2, 2), 'gamma': True},
                      convergence={'density': 1e-7},
                      parallel={'domain': 1},
                      txt='si.txt')
    e = atoms.get_potential_energy()
    scalapack = atoms.calc.wfs.bd.comm.size
    atoms.calc.diagonalize_full_hamiltonian(nbands=8, scalapack=scalapack)
    atoms.calc.write('si.gpw', mode='all')

    # This tests checks the actual numerical accuracy which is asserted below
    gw = G0W0('si.gpw', 'gw_None',
              nbands=8, integrate_gamma=0,
              kpts=[(0, 0, 0), (0.5, 0.5, 0)],  # Gamma, X
              ecut=40, nblocks=nblocks,
              frequencies={'type': 'nonlinear',
                           'domega0': 0.1, 'omegamax': None},
              eta=0.2, relbands=(-1, 2))
    results = gw.calculate()

    G, X = results['eps'][0]
    output = [e, G[0], G[1] - G[0], X[1] - G[0], X[2] - X[1]]
    G, X = results['qp'][0]
    output += [G[0], G[1] - G[0], X[1] - G[0], X[2] - X[1]]

    return output


reference = pytest.approx([-9.253, 5.442, 2.389, 0.403, 0.000,
                           6.261, 3.570, 1.323, 0.001], abs=0.0035)


@pytest.mark.response
@pytest.mark.slow
@pytest.mark.parametrize('si', generate_si_systems())
@pytest.mark.parametrize('symm', [{},
                                  'off',
                                  {'time_reversal': False},
                                  {'point_group': False}])
@pytest.mark.parametrize('nblocks',
                         [x for x in [1, 2, 4, 8] if x <= world.size])
def test_response_gwsi(in_tmp_dir, si, symm, nblocks, scalapack,
                       needs_ase_master):
    assert run(si, symm, nblocks) == reference


@pytest.mark.response
@pytest.mark.ci
@pytest.mark.parametrize('si', generate_si_systems())
@pytest.mark.parametrize('symm', [{}])
def test_small_response_gwsi(in_tmp_dir, si, symm, scalapack,
                             needs_ase_master):
    assert run(si, symm, 1) == reference


@pytest.mark.response
@pytest.mark.ci
def test_few_freq_response_gwsi(in_tmp_dir, scalapack, needs_ase_master):
    if world.size > 1:
        nblocks = 2
    else:
        nblocks = 1

    # This test has very few frequencies and tests that the code doesn't crash.
    atoms = bulk('Si', 'diamond', a=5.43)
    atoms.calc = GPAW(mode=PW(250),
                      eigensolver='rmm-diis',
                      occupations=FermiDirac(0.01),
                      kpts={'size': (2, 2, 2), 'gamma': True},
                      convergence={'density': 1e-7},
                      parallel={'domain': 1},
                      txt='si.txt')
    atoms.get_potential_energy()
    scalapack = atoms.calc.wfs.bd.comm.size
    atoms.calc.diagonalize_full_hamiltonian(nbands=8, scalapack=scalapack)
    atoms.calc.write('si.gpw', mode='all')
    gw = G0W0('si.gpw', 'gw_0.2',
              nbands=8, integrate_gamma=0,
              kpts=[(0, 0, 0), (0.5, 0.5, 0)],  # Gamma, X
              ecut=40, nblocks=nblocks,
              frequencies={'type': 'nonlinear',
                           'domega0': 0.1, 'omegamax': 0.2},
              eta=0.2, relbands=(-1, 2))
    gw.calculate()



from __future__ import annotations
import functools
import os
from time import ctime

import numpy as np
from ase.units import Hartree
from scipy.special import p_roots

import gpaw.mpi as mpi
from gpaw.response import timer
from gpaw.response.chi0 import Chi0Calculator
from gpaw.response.coulomb_kernels import CoulombKernel
from gpaw.response.frequencies import FrequencyDescriptor
from gpaw.response.pair import get_gs_and_context, KPointPairFactory


def default_ecut_extrapolation(ecut, extrapolate):
    return ecut * (1 + 0.5 * np.arange(extrapolate))**(-2 / 3)


def rpa(filename, ecut=200.0, blocks=1, extrapolate=4):
    """Calculate RPA energy.

    filename: str
        Name of restart-file.
    ecut: float
        Plane-wave cutoff.
    blocks: int
        Split polarizability matrix in this many blocks.
    extrapolate: int
        Number of cutoff energies to use for extrapolation.
    """
    name, ext = filename.rsplit('.', 1)
    assert ext == 'gpw'
    from gpaw.xc.rpa import RPACorrelation
    rpa = RPACorrelation(name, name + '-rpa.dat',
                         nblocks=blocks,
                         ecut=default_ecut_extrapolation(ecut, extrapolate),
                         txt=name + '-rpa.txt')
    rpa.calculate()


class GCut:
    def __init__(self, cut_G):
        self._cut_G = cut_G

    @property
    def nG(self):
        return len(self._cut_G)

    def spin_cut(self, array_GG, ns):
        # Strange special case for spin-repeated arrays.
        # Maybe we can get rid of this.
        if self._cut_G is None:
            return array_GG

        cut_sG = np.tile(self._cut_G, ns)
        cut_sG[self.nG:] += len(array_GG) // ns
        array_GG = array_GG.take(cut_sG, 0).take(cut_sG, 1)
        return array_GG

    def cut(self, array, axes=(0,)):
        if self._cut_G is None:
            return array

        for axis in axes:
            array = array.take(self._cut_G, axis)
        return array


def initialize_q_points(kd, qsym):
    bzq_qc = kd.get_bz_q_points(first=True)

    if not qsym:
        ibzq_qc = bzq_qc
        weight_q = np.ones(len(bzq_qc)) / len(bzq_qc)
    else:
        U_scc = kd.symmetry.op_scc
        ibzq_qc = kd.get_ibz_q_points(bzq_qc, U_scc)[0]
        weight_q = kd.q_weights
    return bzq_qc, ibzq_qc, weight_q


class RPACalculator:
    def __init__(self, gs, *, context, filename=None,
                 ecut,
                 skip_gamma=False, qsym=True,
                 frequencies, weights, truncation=None,
                 nblocks=1, calculate_q=None):
        self.gs = gs
        self.context = context

        self.omega_w = frequencies / Hartree
        self.weight_w = weights / Hartree

        # TODO: We should avoid this requirement.
        assert len(self.omega_w) % nblocks == 0

        self.nblocks = nblocks

        self.coulomb = CoulombKernel.from_gs(gs, truncation=truncation)
        self.skip_gamma = skip_gamma

        # We should actually have a kpoint descriptor for the qpoints.
        # We are badly failing at making use of the existing tools by reducing
        # the qpoints to dumb arrays.
        self.bzq_qc, self.ibzq_qc, self.weight_q = initialize_q_points(
            gs.kd, qsym)

        self.filename = filename

        if calculate_q is None:
            calculate_q = self.calculate_q_rpa
        self.calculate_q = calculate_q

        if isinstance(ecut, (float, int)):
            ecut = default_ecut_extrapolation(ecut, extrapolate=6)
        self.ecut_i = np.asarray(np.sort(ecut)) / Hartree

    def read(self, ecut_i, filename):
        with open(filename) as fd:
            lines = fd.readlines()[1:]

        n = 0
        energy_qi = []
        nq = len(lines) // len(ecut_i)
        for q_c in self.ibzq_qc[:nq]:
            energy_qi.append([])
            for ecut in ecut_i:
                current_inputs = np.array([*q_c, ecut * Hartree])
                numbers_from_file = [float(x) for x in lines[n].split()]
                previous_inputs = numbers_from_file[:-1]

                if not np.allclose(current_inputs, previous_inputs):
                    # Energies are not reusable since input parameters
                    # have changed
                    return []

                energy = numbers_from_file[-1]
                energy_qi[-1].append(energy / Hartree)
                n += 1

        return energy_qi

    def energies_to_string(self, energy_qi, ecut_i):
        lines = []
        app = lines.append
        app('q1 q2 q3 E_cut E_c(q)')
        for energy_i, q_c in zip(energy_qi, self.ibzq_qc):
            for energy, ecut in zip(energy_i, ecut_i):
                tokens = [repr(num) for num in
                          (*q_c, ecut * Hartree, energy * Hartree)]
                app(' '.join(tokens))

    def write(self, energy_qi, ecut_i):
        txt = self.energies_to_string(energy_qi, ecut_i)
        if self.context.comm.rank == 0 and self.filename:
            with open(self.filename, 'w') as fd:
                print(txt, file=fd)

    def calculate(self, *, nbands=None, spin=False):
        """Calculate RPA correlation energy for one or several cutoffs.

        ecut: float or list of floats
            Plane-wave cutoff(s) in eV.
        nbands: int
            Number of bands (defaults to number of plane-waves).
        spin: bool
            Separate spin in response function.
            (Only needed for beyond RPA methods that inherit this function).
        """

        p = functools.partial(self.context.print, flush=False)

        ecut_i = self.ecut_i
        ecutmax = max(ecut_i)

        if nbands is None:
            p('Response function bands : Equal to number of plane waves')
        else:
            p('Response function bands : %s' % nbands)
        p('Plane wave cutoffs (eV) :', end='')
        for e in ecut_i:
            p(' {0:.3f}'.format(e * Hartree), end='')
        p()
        p(self.coulomb.description())
        self.context.print('')

        if self.filename and os.path.isfile(self.filename):
            energy_qi = self.read(ecut_i, self.filename)
            self.context.print(
                'Read %d q-points from file: %s\n' % (len(energy_qi)))

            self.context.comm.barrier()

        wd = FrequencyDescriptor(1j * self.omega_w)

        kptpair_factory = KPointPairFactory(
            self.gs,
            context=self.context.with_txt('chi0.txt'),
            nblocks=self.nblocks)

        chi0calc = Chi0Calculator(wd=wd,
                                  kptpair_factory=kptpair_factory,
                                  eta=0.0,
                                  intraband=False,
                                  hilbert=False,
                                  ecut=ecutmax * Hartree)

        self.blockcomm = chi0calc.integrator.blockcomm

        energy_qi = []
        nq = len(energy_qi)

        self.context.timer.start('RPA')

        for q_c in self.ibzq_qc[nq:]:
            if np.allclose(q_c, 0.0) and self.skip_gamma:
                energy_qi.append(len(ecut_i) * [0.0])
                self.write(energy_qi, ecut_i)
                p('Not calculating E_c(q) at Gamma')
                p()
                continue

            chi0_s = [chi0calc.create_chi0(q_c)]
            if spin:
                chi0_s.append(chi0calc.create_chi0(q_c))

            qpd = chi0_s[0].qpd
            nG = qpd.ngmax

            # First not completely filled band:
            m1 = chi0calc.nocc1
            p('# %s  -  %s' % (len(energy_qi), ctime().split()[-2]))
            p('q = [%1.3f %1.3f %1.3f]' % tuple(q_c))

            energy_i = []
            for ecut in ecut_i:
                if ecut == ecutmax:
                    # Nothing to cut away:
                    gcut = GCut(None)
                    m2 = nbands or nG
                else:
                    gcut = GCut(np.arange(nG)[qpd.G2_qG[0] <= 2 * ecut])
                    m2 = gcut.nG

                p('E_cut = %d eV / Bands = %d:' % (ecut * Hartree, m2),
                  end='\n', flush=True)

                energy = self.calculate_q(chi0calc, chi0_s, m1, m2, gcut)

                energy_i.append(energy)
                m1 = m2

                a = 1 / chi0calc.integrator.kncomm.size
                if ecut < ecutmax and a != 1.0:
                    # Chi0 will be summed again over chicomm, so we divide
                    # by its size:
                    for chi0 in chi0_s:
                        chi0.chi0_WgG[:] *= a
                    # if chi0_swxvG is not None:
                    #     chi0_swxvG *= a
                    #     chi0_swvv *= a

            energy_qi.append(energy_i)
            self.write(energy_qi, ecut_i)
            p()

        e_i = np.dot(self.weight_q, np.array(energy_qi))
        p('==========================================================')
        p()
        p('Total correlation energy:')
        for e_cut, e in zip(ecut_i, e_i):
            p('%6.0f:   %6.4f eV' % (e_cut * Hartree, e * Hartree))
        p()

        if len(e_i) > 1:
            self.extrapolate(e_i, ecut_i)

        p('Calculation completed at: ', ctime())
        p()

        self.context.timer.stop('RPA')
        self.context.write_timer()

        return e_i * Hartree

    @timer('chi0(q)')
    def calculate_q_rpa(self, chi0calc, chi0_s,
                        m1, m2, gcut):
        chi0 = chi0_s[0]
        chi0calc.update_chi0(chi0,
                             m1, m2, spins=chi0calc.get_spins())

        self.context.print('E_c(q) = ', end='', flush=False)

        chi0_wGG = chi0.body.copy_array_with_distribution('wGG')

        kd = self.gs.kd
        if not chi0.qpd.kd.gamma:
            e = self.calculate_energy_rpa(chi0.qpd, chi0_wGG, gcut)
            self.context.print('%.3f eV' % (e * Hartree))
        else:
            from gpaw.response.gamma_int import GammaIntegrator
            from gpaw.response.pw_parallelization import Blocks1D

            wblocks = Blocks1D(self.blockcomm, len(self.omega_w))
            gamma_int = GammaIntegrator(
                truncation=self.coulomb.truncation,
                kd=kd,
                qpd=chi0.qpd,
                chi0_wvv=chi0.chi0_Wvv[wblocks.myslice],
                chi0_wxvG=chi0.chi0_WxvG[wblocks.myslice])

            e = 0
            for iqf in range(len(gamma_int.qf_qv)):
                for iw in range(wblocks.nlocal):
                    gamma_int.set_appendages(chi0_wGG[iw], iw, iqf)
                ev = self.calculate_energy_rpa(chi0.qpd, chi0_wGG, gcut,
                                               q_v=gamma_int.qf_qv[iqf])
                e += ev * gamma_int.weight_q
            self.context.print('%.3f eV' % (e * Hartree))

        return e

    @timer('Energy')
    def calculate_energy_rpa(self, qpd, chi0_wGG, gcut, q_v=None):
        """Evaluate correlation energy from chi0."""

        sqrtV_G = gcut.cut(self.coulomb.sqrtV(qpd, q_v))

        nG = len(sqrtV_G)

        e_w = []
        for chi0_GG in chi0_wGG:
            chi0_GG = gcut.cut(chi0_GG, [0, 1])

            e_GG = np.eye(nG) - chi0_GG * sqrtV_G * sqrtV_G[:, np.newaxis]
            e = np.log(np.linalg.det(e_GG)) + nG - np.trace(e_GG)
            e_w.append(e.real)

        self.E_w, energy = self.gather_energies(e_w)
        return energy

    def gather_energies(self, e_w):
        E_w = np.zeros_like(self.omega_w)
        # XXX This requires all cores to the same number of w doesn't it?
        self.blockcomm.all_gather(np.array(e_w), E_w)
        energy = E_w @ self.weight_w / (2 * np.pi)
        return E_w, energy

    def extrapolate(self, e_i, ecut_i):
        self.context.print('Extrapolated energies:', flush=False)
        ex_i = []
        for i in range(len(e_i) - 1):
            e1, e2 = e_i[i:i + 2]
            x1, x2 = ecut_i[i:i + 2]**-1.5
            ex = (e1 * x2 - e2 * x1) / (x2 - x1)
            ex_i.append(ex)

            self.context.print('  %4.0f -%4.0f:  %5.3f eV' %
                               (ecut_i[i] * Hartree, ecut_i[i + 1]
                                * Hartree, ex * Hartree), flush=False)
        self.context.print('')

        return e_i * Hartree


def get_gauss_legendre_points(nw=16, frequency_max=800.0, frequency_scale=2.0):
    y_w, weights_w = p_roots(nw)
    y_w = y_w.real
    ys = 0.5 - 0.5 * y_w
    ys = ys[::-1]
    w = (-np.log(1 - ys))**frequency_scale
    w *= frequency_max / w[-1]
    alpha = (-np.log(1 - ys[-1]))**frequency_scale / frequency_max
    transform = (-np.log(1 - ys))**(frequency_scale - 1) \
        / (1 - ys) * frequency_scale / alpha
    return w, weights_w * transform / 2


class RPACorrelation(RPACalculator):
    def __init__(self, calc, xc='RPA',
                 nlambda=None,
                 nfrequencies=16, frequency_max=800.0, frequency_scale=2.0,
                 frequencies=None, weights=None,
                 world=mpi.world,
                 txt='-',
                 truncation: str | None = None,
                 **kwargs):
        """Creates the RPACorrelation object

        calc: str or calculator object
            The string should refer to the .gpw file contaning KS orbitals
        xc: str
            Exchange-correlation kernel. This is only different from RPA when
            this object is constructed from a different module - e.g. fxc.py
        filename: str
            txt output
        skip_gamme: bool
            If True, skip q = [0,0,0] from the calculation
        qsym: bool
            Use symmetry to reduce q-points
        nlambda: int
            Number of lambda points. Only used for numerical coupling
            constant integration involved when called from fxc.py
        nfrequencies: int
            Number of frequency points used in the Gauss-Legendre integration
        frequency_max: float
            Largest frequency point in Gauss-Legendre integration
        frequency_scale: float
            Determines density of frequency points at low frequencies. A slight
            increase to e.g. 2.5 or 3.0 improves convergence wth respect to
            frequency points for metals
        frequencies: list
            List of frequancies for user-specified frequency integration
        weights: list
            list of weights (integration measure) for a user specified
            frequency grid. Must be specified and have the same length as
            frequencies if frequencies is not None
        truncation: str or None
            Coulomb truncation scheme. Can be None, '0D' or '2D'.  If None
            and the system is a molecule then '0D' will be used.
        world: communicator
        nblocks: int
            Number of parallelization blocks. Frequency parallelization
            can be specified by setting nblocks=nfrequencies and is useful
            for memory consuming calculations
        txt: str
            txt file for saving and loading contributions to the correlation
            energy from different q-points
        """
        gs, context = get_gs_and_context(calc=calc, txt=txt, world=world,
                                         timer=None)

        if frequencies is None:
            frequencies, weights = get_gauss_legendre_points(nfrequencies,
                                                             frequency_max,
                                                             frequency_scale)
            user_spec = False
        else:
            assert weights is not None
            user_spec = True

        if truncation is None and not gs.pbc.any():
            truncation = '0D'

        super().__init__(gs=gs, context=context,
                         frequencies=frequencies, weights=weights,
                         truncation=truncation,
                         **kwargs)

        self.print_initialization(xc, frequency_scale, nlambda, user_spec)

    def print_initialization(self, xc, frequency_scale, nlambda, user_spec):
        p = functools.partial(self.context.print, flush=False)
        p('----------------------------------------------------------')
        p('Non-self-consistent %s correlation energy' % xc)
        p('----------------------------------------------------------')
        p('Started at:  ', ctime())
        p()
        p('Atoms                          :',
          self.gs.atoms.get_chemical_formula(mode='hill'))
        p('Ground state XC functional     :', self.gs.xcname)
        p('Valence electrons              :', self.gs.nvalence)
        p('Number of bands                :', self.gs.bd.nbands)
        p('Number of spins                :', self.gs.nspins)
        p('Number of k-points             :', len(self.gs.kd.bzk_kc))
        p('Number of irreducible k-points :', len(self.gs.kd.ibzk_kc))
        p('Number of q-points             :', len(self.bzq_qc))
        p('Number of irreducible q-points :', len(self.ibzq_qc))
        p()
        for q, weight in zip(self.ibzq_qc, self.weight_q):
            p('    q: [%1.4f %1.4f %1.4f] - weight: %1.3f' %
              (q[0], q[1], q[2], weight))
        p()
        p('----------------------------------------------------------')
        p('----------------------------------------------------------')
        p()
        if nlambda is None:
            p('Analytical coupling constant integration')
        else:
            p('Numerical coupling constant integration using', nlambda,
              'Gauss-Legendre points')
        p()
        p('Frequencies')
        if not user_spec:
            p('    Gauss-Legendre integration with %s frequency points' %
              len(self.omega_w))
            p('    Transformed from [0,oo] to [0,1] using e^[-aw^(1/B)]')
            p('    Highest frequency point at %5.1f eV and B=%1.1f' %
              (self.omega_w[-1] * Hartree, frequency_scale))
        else:
            p('    User specified frequency integration with',
              len(self.omega_w), 'frequency points')
        p()
        p('Parallelization')
        p('    Total number of CPUs          : % s' % self.context.comm.size)
        p('    G-vector decomposition        : % s' % self.nblocks)
        p('    K-point/band decomposition    : % s' %
          (self.context.comm.size // self.nblocks))
        self.context.print('')



import pytest

from ase.build import molecule

from gpaw import GPAW
from gpaw.tddft import TDDFT, DipoleMomentWriter
from gpaw.mpi import world, serial_comm
from gpaw.utilities import compiled_with_sl

from ..lcaotddft.test_molecule import only_on_master


pytestmark = pytest.mark.usefixtures('module_tmp_path')


def calculate_time_propagation(gpw_fpath, *,
                               iterations=3,
                               kick=[1e-5, 1e-5, 1e-5],
                               propagator='SICN',
                               communicator=world,
                               write_and_continue=False,
                               force_new_dm_file=False,
                               parallel={}):
    td_calc = TDDFT(gpw_fpath,
                    propagator=propagator,
                    communicator=communicator,
                    parallel=parallel,
                    txt='td.out')
    DipoleMomentWriter(td_calc, 'dm.dat',
                       force_new_file=force_new_dm_file)
    if kick is not None:
        td_calc.absorption_kick(kick)
    td_calc.propagate(20, iterations)
    if write_and_continue:
        td_calc.write('td.gpw', mode='all')
        # Switch dipole moment writer and output
        td_calc.observers.pop()
        dm = DipoleMomentWriter(td_calc, 'dm2.dat', force_new_file=True)
        dm._update(td_calc)
        td_calc.propagate(20, iterations)
    communicator.barrier()


def check_dm(ref_fpath, fpath, rtol=1e-8, atol=1e-12):
    from gpaw.tddft.spectrum import read_dipole_moment_file

    world.barrier()
    _, time_ref_t, _, dm_ref_tv = read_dipole_moment_file(ref_fpath)
    _, time_t, _, dm_tv = read_dipole_moment_file(fpath)
    assert time_t == pytest.approx(time_ref_t, abs=0)
    assert dm_tv == pytest.approx(dm_ref_tv, rel=rtol, abs=atol)


# Generate different parallelization options
parallel_i = [{}]
if world.size > 1:
    parallel_i.append({'band': 2})
if compiled_with_sl():
    parallel_i.append({'sl_auto': True})
    if world.size > 1:
        parallel_i.append({'sl_auto': True, 'band': 2})


@pytest.fixture(scope='module')
@only_on_master(world)
def ground_state():
    atoms = molecule('SiH4')
    atoms.center(vacuum=4.0)

    calc = GPAW(mode='fd', nbands=6, h=0.4,
                convergence={'density': 1e-8},
                communicator=serial_comm,
                xc='LDA',
                symmetry={'point_group': False},
                txt='gs.out')
    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('gs.gpw', mode='all')


@pytest.fixture(scope='module')
@only_on_master(world)
def time_propagation_reference(ground_state):
    calculate_time_propagation('gs.gpw',
                               communicator=serial_comm,
                               write_and_continue=True)


@pytest.mark.later
def test_dipole_moment_values(time_propagation_reference,
                              module_tmp_path, in_tmp_dir):
    with open('dm.dat', 'w') as fd:
        fd.write('''
# DipoleMomentWriter[version=1](center=False, density='comp')
#            time            norm                    dmx                    dmy                    dmz
# Start; Time = 0.00000000
          0.00000000      -8.62679509e-16     8.856042552837e-09     5.230011358635e-11     1.624559936066e-10
# Kick = [    1.000000000000e-05,     1.000000000000e-05,     1.000000000000e-05]; Time = 0.00000000
          0.00000000       9.59295128e-16     8.826542661185e-09    -1.968118480737e-10    -1.260104338852e-10
          0.82682747       1.64702342e-15     6.016062457419e-05     6.015263997632e-05     6.015070820074e-05
          1.65365493       1.36035859e-15     1.075409609786e-04     1.075366083805e-04     1.075339737111e-04
          2.48048240       1.53109134e-15     1.388608139179e-04     1.388701618472e-04     1.388666380740e-04
'''.strip())  # noqa: E501

    with open('dm2.dat', 'w') as fd:
        fd.write('''
# DipoleMomentWriter[version=1](center=False, density='comp')
#            time            norm                    dmx                    dmy                    dmz
          2.48048240       1.53109134e-15     1.388608139179e-04     1.388701618472e-04     1.388666380740e-04
          3.30730987       1.36214053e-15     1.528275514998e-04     1.528424797241e-04     1.528388409079e-04
          4.13413733      -5.46885441e-16     1.498039918400e-04     1.498178744836e-04     1.498147055362e-04
          4.96096480      -3.62630566e-16     1.324275745486e-04     1.324479404917e-04     1.324450415780e-04
'''.strip())  # noqa: E501

    rtol = 5e-4
    atol = 1e-7
    check_dm('dm.dat', module_tmp_path / 'dm.dat', rtol=rtol, atol=atol)
    check_dm('dm2.dat', module_tmp_path / 'dm2.dat', rtol=rtol, atol=atol)


@pytest.mark.later
@pytest.mark.parametrize('parallel', parallel_i)
@pytest.mark.parametrize('propagator', [
    'SICN', 'ECN', 'ETRSCN', 'SIKE'])
def test_propagation(time_propagation_reference,
                     parallel, propagator,
                     module_tmp_path, in_tmp_dir):
    calculate_time_propagation(module_tmp_path / 'gs.gpw',
                               propagator=propagator,
                               parallel=parallel)
    atol = 1e-12
    if propagator == 'SICN':
        # This is the same propagator as the reference;
        # error comes only from parallelization
        rtol = 1e-8
        if 'band' in parallel:
            # XXX band parallelization is inaccurate!
            rtol = 7e-4
            atol = 5e-8
    else:
        # Other propagators match qualitatively
        rtol = 5e-2
        if 'band' in parallel:
            # XXX band parallelization is inaccurate!
            atol = 5e-8
    check_dm(module_tmp_path / 'dm.dat', 'dm.dat', rtol=rtol, atol=atol)


@pytest.mark.later
@pytest.mark.parametrize('parallel', parallel_i)
def test_restart(time_propagation_reference,
                 parallel,
                 module_tmp_path, in_tmp_dir):
    calculate_time_propagation(module_tmp_path / 'td.gpw',
                               kick=None,
                               force_new_dm_file=True,
                               parallel=parallel)
    rtol = 1e-8
    if 'band' in parallel:
        rtol = 5e-4
    check_dm(module_tmp_path / 'dm2.dat', 'dm.dat', rtol=rtol)



import numpy as np
from gpaw.core import PWDesc
from gpaw.gpu import cupy as cp, T
from gpaw.new.pw.hamiltonian import precondition
import pytest


@pytest.mark.gpu
@pytest.mark.parametrize('xp', [np, cp])
def test_prec(xp):
    a = 2
    pw = PWDesc(cell=[a, a, a], ecut=200 / 27, dtype=complex)
    n = 2
    psit_nG, residual_nG, out_nG = pw.zeros((3, n), xp=xp)
    psit_nG.data[:, :2] = 1.0
    residual_nG.data[:] = 1.0
    print(residual_nG)
    for _ in range(1):
        with T():
            precondition(psit_nG, residual_nG, out_nG)



import pytest
from itertools import count
from ase.build import bulk
from gpaw import GPAW
from gpaw.mpi import world
from gpaw.utilities import compiled_with_sl


def ikwargs():
    for augment_grids in [False, True]:
        for sl_auto in ([0, 1] if compiled_with_sl() else [0]):
            if world.size == 16:
                # This won't happen in ordinary test suite
                parsizes = [[4, 2, 2], [2, 4, 2], [2, 2, 4]]
            elif world.size == 8:
                parsizes = [[2, 2, 2]]
            elif world.size == 4:
                parsizes = [[1, 2, 2]]
            elif world.size == 2:
                parsizes = [[1, 1, 2]]
            else:
                assert world.size == 1
                parsizes = [[1, 1, 1]]
            for kpt, band, domain in parsizes:
                parallel = dict(kpt=kpt,
                                band=band,
                                domain=domain,
                                sl_auto=sl_auto,
                                augment_grids=augment_grids)
                yield dict(parallel=parallel)


@pytest.mark.later
def test_lcao_kpts_many_combinations(in_tmp_dir):
    counter = count()

    for spinpol in [False, True]:
        # We want a non-trivial cell:
        atoms0 = bulk('Ti') * (2, 1, 1)
        atoms0.cell[0] *= 1.2
        # We want most arrays to be different so we can detect ordering/shape
        # trouble:
        atoms0.symbols = 'HOFePb'
        atoms0.rattle(stdev=0.1)

        if spinpol:
            atoms0.set_initial_magnetic_moments([1.0] * len(atoms0))

        kwargs = []
        energies = []
        eerrs = []
        forces = []
        ferrs = []

        from time import time
        for kwargs in ikwargs():
            i = next(counter)

            if world.rank == 0:
                print(i, kwargs)

            if kwargs['parallel']['kpt'] == 4 and not spinpol:
                continue  # Core without kpoints

            calc = GPAW(
                mode='lcao',
                basis='sz(dzp)',
                xc='PBE', h=0.3,
                symmetry={'point_group': False},  # No symmetry here anyway
                txt='gpaw.{:02d}.spin{}.txt'.format(int(spinpol), i),
                kpts=(4, 1, 1),
                convergence={'maximum iterations': 2},
                **kwargs)
            atoms = atoms0.copy()
            t1 = time()
            atoms.calc = calc
            e = atoms.get_potential_energy()
            f = atoms.get_forces()
            t2 = time()
            if world.rank == 0:
                print('T', t2 - t1)
            energies.append(e)
            forces.append(f)
            corrname = calc.wfs.atomic_correction.name
            if kwargs['parallel']['sl_auto']:
                assert corrname == 'sparse'
            else:
                assert corrname == 'dense'

            if energies:
                eerr = abs(e - energies[0])
                ferr = abs(f - forces[0]).max()
                eerrs.append(eerr)
                ferrs.append(ferr)
                if world.rank == 0:
                    print('Eerr {} Ferr {}'.format(eerr, ferr))
                    print()

        if world.rank == 0:
            print('eerrs', eerrs)
            print('ferrs', ferrs)

            maxeerr = max(eerrs)
            maxferr = max(ferrs)
            print('maxeerr', maxeerr)
            print('maxferr', maxferr)
            assert maxeerr < 1e-9, maxeerr
            assert maxferr < 1e-10, maxferr



import sys
from pathlib import Path

from ase.cli.completion import update, CLICommand

from gpaw.cli.main import commands

# Path of the complete.py script:
path = Path(__file__).with_name('complete.py')

CLICommand.cmd = f'complete -o default -C "{sys.executable} {path}" gpaw'

if __name__ == '__main__':
    update(path, commands)



from math import pi

import _gpaw
import numpy as np
from gpaw.lfc import BaseLFC
from gpaw.spherical_harmonics import Y, nablarlYL
from gpaw.ffbt import rescaled_fourier_bessel_transform
from gpaw.utilities.blas import mmm


class PWLFC(BaseLFC):
    def __init__(self, spline_aj, pd, blocksize=5000, comm=None):
        """Reciprocal-space plane-wave localized function collection.

        spline_aj: list of list of spline objects
            Splines.
        pd: PWDescriptor
            Plane-wave descriptor object.
        blocksize: int
            Block-size to use when looping over G-vectors.  Use None for
            doing all G-vectors in one big block.
        comm: communicator
            Communicator for operations that support parallelization
            over planewaves (only integrate so far)."""

        self.pd = pd
        self.spline_aj = spline_aj

        self.dtype = pd.dtype

        self.initialized = False

        # These will be filled in later:
        self.Y_qGL = []
        self.emiGR_qGa = []
        self.f_qGs = []
        self.l_s = None
        self.a_J = None
        self.s_J = None
        self.lmax = None

        if blocksize is not None:
            if pd.ngmax <= blocksize:
                # No need to block G-vectors
                blocksize = None
        self.blocksize = blocksize

        # These are set later in set_potitions():
        self.eikR_qa = None
        self.my_atom_indices = None
        self.my_indices = None
        self.pos_av = None
        self.nI = None

        if comm is None:
            comm = pd.gd.comm
        else:
            assert False
        self.comm = comm

    def initialize(self):
        """Initialize position-independent stuff."""
        if self.initialized:
            return

        splines = {}  # Dict[Spline, int]
        for spline_j in self.spline_aj:
            for spline in spline_j:
                if spline not in splines:
                    splines[spline] = len(splines)
        nsplines = len(splines)

        nJ = sum(len(spline_j) for spline_j in self.spline_aj)

        self.f_qGs = [np.empty((mynG, nsplines)) for mynG in self.pd.myng_q]
        self.l_s = np.empty(nsplines, np.int32)
        self.a_J = np.empty(nJ, np.int32)
        self.s_J = np.empty(nJ, np.int32)

        # Fourier transform radial functions:
        J = 0
        done = set()  # Set[Spline]
        for a, spline_j in enumerate(self.spline_aj):
            for spline in spline_j:
                s = splines[spline]  # get spline index
                if spline not in done:
                    f = rescaled_fourier_bessel_transform(spline)
                    for f_Gs, G2_G in zip(self.f_qGs, self.pd.G2_qG):
                        G_G = G2_G**0.5
                        f_Gs[:, s] = f.map(G_G)
                    self.l_s[s] = spline.get_angular_momentum_number()
                    done.add(spline)
                self.a_J[J] = a
                self.s_J[J] = s
                J += 1

        self.lmax = max(self.l_s, default=-1)

        # Spherical harmonics:
        for q, K_v in enumerate(self.pd.K_qv):
            G_Gv = self.pd.get_reciprocal_vectors(q=q)
            Y_GL = np.empty((len(G_Gv), (self.lmax + 1)**2))
            for L in range((self.lmax + 1)**2):
                Y_GL[:, L] = Y(L, *G_Gv.T)
            self.Y_qGL.append(Y_GL)

        self.initialized = True

    def estimate_memory(self, mem):
        splines = set()
        lmax = -1
        for spline_j in self.spline_aj:
            for spline in spline_j:
                splines.add(spline)
                l = spline.get_angular_momentum_number()
                lmax = max(lmax, l)
        nbytes = ((len(splines) + (lmax + 1)**2) *
                  sum(G2_G.nbytes for G2_G in self.pd.G2_qG))
        mem.subnode('Arrays', nbytes)

    def get_function_count(self, a):
        return sum(2 * spline.get_angular_momentum_number() + 1
                   for spline in self.spline_aj[a])

    def set_positions(self, spos_ac, atom_partition=None):
        self.initialize()
        kd = self.pd.kd
        if kd is None or kd.gamma:
            self.eikR_qa = np.ones((1, len(spos_ac)))
        else:
            self.eikR_qa = np.exp(2j * pi * np.dot(kd.ibzk_qc, spos_ac.T))

        self.pos_av = np.dot(spos_ac, self.pd.gd.cell_cv)

        del self.emiGR_qGa[:]
        G_Qv = self.pd.G_Qv
        for Q_G in self.pd.myQ_qG:
            GR_Ga = np.dot(G_Qv[Q_G], self.pos_av.T)
            self.emiGR_qGa.append(np.exp(-1j * GR_Ga))

        if atom_partition is None:
            assert self.comm.size == 1
            rank_a = np.zeros(len(spos_ac), int)
        else:
            rank_a = atom_partition.rank_a

        self.my_atom_indices = []
        self.my_indices = []
        I1 = 0
        for a, rank in enumerate(rank_a):
            I2 = I1 + self.get_function_count(a)
            if rank == self.comm.rank:
                self.my_atom_indices.append(a)
                self.my_indices.append((a, I1, I2))
            I1 = I2
        self.nI = I1

    def expand(self, q=-1, G1=0, G2=None, cc=False):
        """Expand functions in plane-waves.

        q: int
            k-point index.
        G1: int
            Start G-vector index.
        G2: int
            End G-vector index.
        cc: bool
            Complex conjugate.
        """
        if G2 is None:
            G2 = self.Y_qGL[q].shape[0]

        emiGR_Ga = self.emiGR_qGa[q][G1:G2]
        f_Gs = self.f_qGs[q][G1:G2]
        Y_GL = self.Y_qGL[q][G1:G2]

        if self.pd.dtype == complex:
            f_GI = np.empty((G2 - G1, self.nI), complex)
        else:
            # Special layout because BLAS does not have real-complex
            # multiplications.  f_GI(G,I) layout:
            #
            #    real(G1, 0),   real(G1, 1),   ...
            #    imag(G1, 0),   imag(G1, 1),   ...
            #    real(G1+1, 0), real(G1+1, 1), ...
            #    imag(G1+1, 0), imag(G1+1, 1), ...
            #    ...

            f_GI = np.empty((2 * (G2 - G1), self.nI))

        if True:
            # Fast C-code:
            _gpaw.pwlfc_expand(f_Gs, emiGR_Ga, Y_GL,
                               self.l_s, self.a_J, self.s_J,
                               cc, f_GI)
            return f_GI

        # Equivalent slow Python code:
        f_GI = np.empty((G2 - G1, self.nI), complex)
        I1 = 0
        for J, (a, s) in enumerate(zip(self.a_J, self.s_J)):
            l = self.l_s[s]
            I2 = I1 + 2 * l + 1
            f_GI[:, I1:I2] = (f_Gs[:, s] *
                              emiGR_Ga[:, a] *
                              Y_GL[:, l**2:(l + 1)**2].T *
                              (-1.0j)**l).T
            I1 = I2
        if cc:
            f_GI = f_GI.conj()
        if self.pd.dtype == float:
            f_GI = f_GI.T.copy().view(float).T.copy()

        return f_GI

    def block(self, q=-1, ensure_same_number_of_blocks=False):
        nG = self.Y_qGL[q].shape[0]
        B = self.blocksize
        if B:
            G1 = 0
            while G1 < nG:
                G2 = min(G1 + B, nG)
                yield G1, G2
                G1 = G2
            if ensure_same_number_of_blocks:
                # Make sure we yield the same number of times:
                nb = (self.pd.maxmyng + B - 1) // B
                mynb = (nG + B - 1) // B
                if mynb < nb:
                    yield nG, nG  # empty block
        else:
            yield 0, nG

    def add(self, a_xG, c_axi=1.0, q=-1, f0_IG=None):
        c_xI = np.empty(a_xG.shape[:-1] + (self.nI,), self.pd.dtype)

        if isinstance(c_axi, float):
            assert q == -1 and a_xG.ndim == 1
            c_xI[:] = c_axi
        else:
            assert q != -1 or self.pd.only_one_k_point
            if self.comm.size != 1:
                c_xI[:] = 0.0
            for a, I1, I2 in self.my_indices:
                c_xI[..., I1:I2] = c_axi[a] * self.eikR_qa[q][a].conj()
            if self.comm.size != 1:
                self.comm.sum(c_xI)

        nx = np.prod(c_xI.shape[:-1], dtype=int)
        c_xI = c_xI.reshape((nx, self.nI))
        a_xG = a_xG.reshape((nx, a_xG.shape[-1])).view(self.pd.dtype)

        for G1, G2 in self.block(q):
            if f0_IG is None:
                f_GI = self.expand(q, G1, G2, cc=False)
            else:
                1 / 0
                # f_IG = f0_IG

            if self.pd.dtype == float:
                # f_IG = f_IG.view(float)
                G1 *= 2
                G2 *= 2

            mmm(1.0 / self.pd.gd.dv, c_xI, 'N', f_GI, 'T',
                1.0, a_xG[:, G1:G2])

    def integrate(self, a_xG, c_axi=None, q=-1):
        c_xI = np.zeros(a_xG.shape[:-1] + (self.nI,), self.pd.dtype)

        nx = np.prod(c_xI.shape[:-1], dtype=int)
        b_xI = c_xI.reshape((nx, self.nI))
        a_xG = a_xG.reshape((nx, a_xG.shape[-1]))

        alpha = 1.0 / self.pd.gd.N_c.prod()
        if self.pd.dtype == float:
            alpha *= 2
            a_xG = a_xG.view(float)

        if c_axi is None:
            c_axi = self.dict(a_xG.shape[:-1])

        x = 0.0
        for G1, G2 in self.block(q):
            f_GI = self.expand(q, G1, G2, cc=self.pd.dtype == complex)
            if self.pd.dtype == float:
                if G1 == 0 and self.comm.rank == 0:
                    f_GI[0] *= 0.5
                G1 *= 2
                G2 *= 2
            mmm(alpha, a_xG[:, G1:G2], 'N', f_GI, 'N', x, b_xI)
            x = 1.0

        self.comm.sum(b_xI)
        for a, I1, I2 in self.my_indices:
            c_axi[a][:] = self.eikR_qa[q][a] * c_xI[..., I1:I2]

        return c_axi

    def matrix_elements(self, psit, out):
        P_ani = {a: P_in.T for a, P_in in out.items()}
        self.integrate(psit.array, P_ani, psit.kpt)

    def derivative(self, a_xG, c_axiv=None, q=-1):
        c_vxI = np.zeros((3,) + a_xG.shape[:-1] + (self.nI,), self.pd.dtype)
        nx = np.prod(c_vxI.shape[1:-1], dtype=int)
        b_vxI = c_vxI.reshape((3, nx, self.nI))
        a_xG = a_xG.reshape((nx, a_xG.shape[-1])).view(self.pd.dtype)

        alpha = 1.0 / self.pd.gd.N_c.prod()

        if c_axiv is None:
            c_axiv = self.dict(a_xG.shape[:-1], derivative=True)

        K_v = self.pd.K_qv[q]

        x = 0.0
        for G1, G2 in self.block(q):
            f_GI = self.expand(q, G1, G2, cc=True)
            G_Gv = self.pd.G_Qv[self.pd.myQ_qG[q][G1:G2]]
            if self.pd.dtype == float:
                d_GI = np.empty_like(f_GI)
                for v in range(3):
                    d_GI[::2] = f_GI[1::2] * G_Gv[:, v, np.newaxis]
                    d_GI[1::2] = f_GI[::2] * G_Gv[:, v, np.newaxis]
                    mmm(2 * alpha,
                        a_xG[:, 2 * G1:2 * G2], 'N',
                        d_GI, 'N',
                        x, b_vxI[v])
            else:
                for v in range(3):
                    mmm(-alpha,
                        a_xG[:, G1:G2], 'N',
                        f_GI * (G_Gv[:, v] + K_v[v])[:, np.newaxis], 'N',
                        x, b_vxI[v])
            x = 1.0

        self.comm.sum(c_vxI)

        for v in range(3):
            if self.pd.dtype == float:
                for a, I1, I2 in self.my_indices:
                    c_axiv[a][..., v] = c_vxI[v, ..., I1:I2]
            else:
                for a, I1, I2 in self.my_indices:
                    c_axiv[a][..., v] = (1.0j * self.eikR_qa[q][a] *
                                         c_vxI[v, ..., I1:I2])

        return c_axiv

    def stress_tensor_contribution(self, a_xG, c_axi=1.0, q=-1):
        cache = {}
        things = []
        I1 = 0
        lmax = 0
        for a, spline_j in enumerate(self.spline_aj):
            for spline in spline_j:
                if spline not in cache:
                    s = rescaled_fourier_bessel_transform(spline)
                    G_G = self.pd.G2_qG[q]**0.5
                    f_G = []
                    dfdGoG_G = []
                    for G in G_G:
                        f, dfdG = s.get_value_and_derivative(G)
                        if G < 1e-10:
                            G = 1.0
                        f_G.append(f)
                        dfdGoG_G.append(dfdG / G)
                    f_G = np.array(f_G)
                    dfdGoG_G = np.array(dfdGoG_G)
                    cache[spline] = (f_G, dfdGoG_G)
                else:
                    f_G, dfdGoG_G = cache[spline]
                l = spline.l
                lmax = max(l, lmax)
                I2 = I1 + 2 * l + 1
                things.append((a, l, I1, I2, f_G, dfdGoG_G))
                I1 = I2

        if isinstance(c_axi, float):
            c_axi = dict((a, c_axi) for a in range(len(self.pos_av)))

        G0_Gv = self.pd.get_reciprocal_vectors(q=q)

        stress_vv = np.zeros((3, 3))
        for G1, G2 in self.block(q, ensure_same_number_of_blocks=True):
            G_Gv = G0_Gv[G1:G2]
            Z_LvG = np.array([nablarlYL(L, G_Gv.T)
                              for L in range((lmax + 1)**2)])
            aa_xG = a_xG[..., G1:G2]
            for v1 in range(3):
                for v2 in range(3):
                    stress_vv[v1, v2] += self._stress_tensor_contribution(
                        v1, v2, things, G1, G2, G_Gv, aa_xG, c_axi, q, Z_LvG)

        self.comm.sum(stress_vv)

        return stress_vv

    def _stress_tensor_contribution(self, v1, v2, things, G1, G2,
                                    G_Gv, a_xG, c_axi, q, Z_LvG):
        f_IG = np.empty((self.nI, G2 - G1), complex)
        emiGR_Ga = self.emiGR_qGa[q][G1:G2]
        Y_LG = self.Y_qGL[q].T
        for a, l, I1, I2, f_G, dfdGoG_G in things:
            L1 = l**2
            L2 = (l + 1)**2
            f_IG[I1:I2] = (emiGR_Ga[:, a] * (-1.0j)**l *
                           (dfdGoG_G[G1:G2] * G_Gv[:, v1] * G_Gv[:, v2] *
                            Y_LG[L1:L2, G1:G2] +
                            f_G[G1:G2] * G_Gv[:, v1] * Z_LvG[L1:L2, v2]))

        c_xI = np.zeros(a_xG.shape[:-1] + (self.nI,), self.pd.dtype)

        x = np.prod(c_xI.shape[:-1], dtype=int)
        b_xI = c_xI.reshape((x, self.nI))
        a_xG = a_xG.reshape((x, a_xG.shape[-1]))

        alpha = 1.0 / self.pd.gd.N_c.prod()
        if self.pd.dtype == float:
            alpha *= 2
            if G1 == 0 and self.pd.gd.comm.rank == 0:
                f_IG[:, 0] *= 0.5
            f_IG = f_IG.view(float)
            a_xG = a_xG.copy().view(float)

        mmm(alpha, a_xG, 'N', f_IG, 'C', 0.0, b_xI)
        self.comm.sum(b_xI)

        stress = 0.0
        for a, I1, I2 in self.my_indices:
            stress -= self.eikR_qa[q][a] * (c_axi[a] * c_xI[..., I1:I2]).sum()
        return stress.real



import pytest
import numpy as np

from ase import Atoms
from gpaw import GPAW, PW
from gpaw.mpi import world
from gpaw.test import equal, findpeak
from gpaw.utilities import compiled_with_sl
from gpaw.response.df import DielectricFunction

# Comparing the plasmon peaks found in bulk sodium for two different
# atomic structures. Testing for idential plasmon peaks. Not using
# physical sodium cell.


@pytest.mark.response
def test_response_na_plasmon(in_tmp_dir):
    a = 4.23 / 2.0
    a1 = Atoms('Na',
               scaled_positions=[[0, 0.1, 0]],
               cell=(a, a, a),
               pbc=True)

    # Expanding along x-direction
    a2 = Atoms('Na2',
               scaled_positions=[[0, 0.1, 0], [0.5, 0.1, 0]],
               cell=(2 * a, a, a),
               pbc=True)

    a1.calc = GPAW(mode=PW(250),
                   kpts={'size': (8, 8, 8), 'gamma': True},
                   parallel={'band': 1},
                   # txt='small.txt',
                   )

    # Kpoint sampling should be halved in the expanded direction.
    a2.calc = GPAW(mode=PW(250),
                   kpts={'size': (4, 8, 8), 'gamma': True},
                   parallel={'band': 1},
                   # txt='large.txt',
                   )

    a1.get_potential_energy()
    a2.get_potential_energy()

    # Use twice as many bands for expanded structure
    a1.calc.diagonalize_full_hamiltonian(nbands=20)
    a2.calc.diagonalize_full_hamiltonian(nbands=40)

    a1.calc.write('gs_Na_small.gpw', 'all')
    a2.calc.write('gs_Na_large.gpw', 'all')

    # Settings that should yield the same result
    settings = [{'disable_point_group': True, 'disable_time_reversal': True},
                {'disable_point_group': False, 'disable_time_reversal': True},
                {'disable_point_group': True, 'disable_time_reversal': False},
                {'disable_point_group': False, 'disable_time_reversal': False}]

    # Test block parallelization (needs scalapack)
    if world.size > 1 and compiled_with_sl():
        settings.append({'disable_point_group': False,
                         'disable_time_reversal': False,
                         'nblocks': 2})

    # Calculate the dielectric functions
    dfs0 = []  # Arrays to check for self-consistency
    dfs1 = []
    dfs2 = []
    dfs3 = []
    dfs4 = []
    dfs5 = []
    for kwargs in settings:
        df1 = DielectricFunction('gs_Na_small.gpw',
                                 ecut=40,
                                 rate=0.001,
                                 **kwargs)

        df1NLFCx, df1LFCx = df1.get_dielectric_function(direction='x')
        df1NLFCy, df1LFCy = df1.get_dielectric_function(direction='y')
        df1NLFCz, df1LFCz = df1.get_dielectric_function(direction='z')

        df2 = DielectricFunction('gs_Na_large.gpw',
                                 ecut=40,
                                 rate=0.001,
                                 **kwargs)

        df2NLFCx, df2LFCx = df2.get_dielectric_function(direction='x')
        df2NLFCy, df2LFCy = df2.get_dielectric_function(direction='y')
        df2NLFCz, df2LFCz = df2.get_dielectric_function(direction='z')

        dfs0.append(df1NLFCx)
        dfs1.append(df1LFCx)
        dfs2.append(df1NLFCy)
        dfs3.append(df1LFCy)
        dfs4.append(df1NLFCz)
        dfs5.append(df1LFCz)

        # Compare plasmon frequencies and intensities
        w_w = df1.wd.omega_w
        w1, I1 = findpeak(w_w, -(1. / df1LFCx).imag)
        w2, I2 = findpeak(w_w, -(1. / df2LFCx).imag)
        equal(w1, w2, 1e-2)
        equal(I1, I2, 1e-3)

        w1, I1 = findpeak(w_w, -(1. / df1LFCy).imag)
        w2, I2 = findpeak(w_w, -(1. / df2LFCy).imag)
        equal(w1, w2, 1e-2)
        equal(I1, I2, 1e-3)

        w1, I1 = findpeak(w_w, -(1. / df1LFCz).imag)
        w2, I2 = findpeak(w_w, -(1. / df2LFCz).imag)
        equal(w1, w2, 1e-2)
        equal(I1, I2, 1e-3)

    # Check for self-consistency
    for i, dfs in enumerate([dfs0, dfs1, dfs2, dfs3, dfs4, dfs5]):
        while len(dfs):
            df = dfs.pop()
            for df2 in dfs:
                try:
                    assert np.max(np.abs((df - df2) / df)) < 1e-3
                except AssertionError:
                    print(np.max(np.abs((df - df2) / df)))
                    raise AssertionError



"""The gpaw.solvation package.

This packages extends GPAW to be used with different
continuum solvent models.
"""

from gpaw.solvation.calculator import SolvationGPAW
from gpaw.solvation.cavity import (EffectivePotentialCavity,
                                   Power12Potential,
                                   ElDensity,
                                   SSS09Density,
                                   ADM12SmoothStepCavity,
                                   FG02SmoothStepCavity,
                                   GradientSurface,
                                   KB51Volume)
from gpaw.solvation.dielectric import (LinearDielectric,
                                       CMDielectric)
from gpaw.solvation.interactions import (SurfaceInteraction,
                                         VolumeInteraction,
                                         LeakedDensityInteraction)


def get_HW14_water_kwargs():
    """Return kwargs for initializing a SolvationGPAW instance.

    Parameters for water as a solvent as in
    A. Held and M. Walter, J. Chem. Phys. 141, 174108 (2014).
    """
    from ase.units import Pascal, m
    from ase.data.vdw import vdw_radii
    u0 = 0.180
    epsinf = 78.36
    st = 18.4 * 1e-3 * Pascal * m
    T = 298.15
    vdw_radii = vdw_radii.copy()
    vdw_radii[1] = 1.09

    def atomic_radii(atoms):
        return [vdw_radii[n] for n in atoms.numbers]

    kwargs = {
        'cavity': EffectivePotentialCavity(
            effective_potential=Power12Potential(atomic_radii, u0),
            temperature=T,
            surface_calculator=GradientSurface()
        ),
        'dielectric': LinearDielectric(epsinf=epsinf),
        'interactions': [SurfaceInteraction(surface_tension=st)]
    }
    return kwargs


__all__ = ['SolvationGPAW',
           'EffectivePotentialCavity',
           'Power12Potential',
           'ElDensity',
           'SSS09Density',
           'ADM12SmoothStepCavity',
           'FG02SmoothStepCavity',
           'GradientSurface',
           'KB51Volume',
           'LinearDielectric',
           'CMDielectric',
           'SurfaceInteraction',
           'VolumeInteraction',
           'LeakedDensityInteraction',
           'get_HW14_water_kwargs']



"""Hyperfine parameters.

See:

    First-principles calculations of defects in oxygen-deficient
    silica exposed to hydrogen

    Peter E. Blöchl

    Phys. Rev. B 62, 6158 – Published 1 September 2000

    https://doi.org/10.1103/PhysRevB.62.6158

"""
import argparse
from math import pi
from typing import List

import ase.units as units
import numpy as np
from scipy.integrate import simpson

from gpaw import SCIPY_VERSION
from gpaw.atom.aeatom import Channel
from gpaw.atom.configurations import configurations
from gpaw.atom.radialgd import RadialGridDescriptor
from gpaw.calculator import GPAW
from gpaw.gaunt import gaunt
from gpaw.grid_descriptor import GridDescriptor
from gpaw.pw.descriptor import PWDescriptor
from gpaw.setup import Setup
from gpaw.typing import Array1D, Array2D, Array3D
from gpaw.utilities import unpack2
from gpaw.xc.functional import XCFunctional

# Fine-structure constant: (~1/137)
alpha = 0.5 * units._mu0 * units._c * units._e**2 / units._hplanck

G_FACTOR_E = 2.00231930436256


def hyperfine_parameters(calc: GPAW,
                         exclude_core=False) -> Array3D:
    r"""Calculate isotropic and anisotropic hyperfine coupling parameters.

    One tensor (:math:`A_{ij}`) per atom is returned in eV units.
    In Hartree atomic units, we have the isotropic part
    :math:`a = \text{Tr}(\mathbf{A}) / 3`:

    .. math::

        a = \frac{2 \alpha^2 g_e m_e}{3 m_p}
            \int \delta_T(\mathbf{r}) \rho_s(\mathbf{r}) d\mathbf{r},

    and the anisotropic part :math:`\mathbf{A} - a`:

    .. math::

        \frac{\alpha^2 g_e m_e}{4 \pi m_p}
        \int \frac{3 r_i r_j - \delta_{ij} r^2}{r^5}
        \rho_s(\mathbf{r}) d\mathbf{r}.

    Remember to multiply each tensor by the g-factors of the nuclei.

    Use ``exclude_core=True`` to exclude contribution from "frozen" core.
    """
    dens = calc.density
    nt_sR = dens.nt_sG
    A_avv = smooth_part(
        nt_sR[0] - nt_sR[1],
        dens.gd,
        calc.atoms.get_scaled_positions())

    D_asp = calc.density.D_asp
    for a, D_sp in D_asp.items():
        density_sii = unpack2(D_sp)
        setup = calc.wfs.setups[a]

        A_vv = paw_correction(density_sii,
                              setup,
                              calc.hamiltonian.xc,
                              exclude_core)

        A_avv[a] += A_vv

    A_avv *= pi * alpha**2 * G_FACTOR_E * units._me / units._mp * units.Ha

    return A_avv


def smooth_part(spin_density_R: Array3D,
                gd: GridDescriptor,
                spos_ac: Array2D,
                ecut: float = None) -> Array3D:
    """Contribution from pseudo spin-density."""
    pd = PWDescriptor(ecut, gd)
    spin_density_G = pd.fft(spin_density_R)
    G_Gv = pd.get_reciprocal_vectors(add_q=False)
    # eiGR_aG = np.exp(-1j * spos_ac.dot(gd.cell_cv).dot(G_Gv.T))
    eiGR_aG = np.exp(-1j * spos_ac @ gd.cell_cv @ G_Gv.T)

    # Isotropic term:
    W1_a = pd.integrate(spin_density_G, eiGR_aG) / gd.dv * (2 / 3)

    spin_density_G[0] = 0.0
    G2_G = pd.G2_qG[0].copy()
    G2_G[0] = 1.0
    spin_density_G /= G2_G

    # Anisotropic term:
    W_vva = np.empty((3, 3, len(spos_ac)))
    for v1 in range(3):
        for v2 in range(3):
            W_a = pd.integrate(G_Gv[:, v1] * G_Gv[:, v2] * spin_density_G,
                               eiGR_aG)
            W_vva[v1, v2] = -W_a / gd.dv

    W_a = np.trace(W_vva) / 3
    for v in range(3):
        W_vva[v, v] -= W_a
        W_vva[v, v] += W1_a

    return W_vva.transpose((2, 0, 1))


# Normalization constants for xy, yz, 3z^2-r^2, xz, x^2-y^2:
Y2_m = (np.array([15 / 4,
                  15 / 4,
                  5 / 16,
                  15 / 4,
                  15 / 16]) / pi)**0.5
# Second derivatives:
Y2_mvv = np.array([[[0, 1, 0],
                    [1, 0, 0],
                    [0, 0, 0]],
                   [[0, 0, 0],
                    [0, 0, 1],
                    [0, 1, 0]],
                   [[-2, 0, 0],
                    [0, -2, 0],
                    [0, 0, 4]],
                   [[0, 0, 1],
                    [0, 0, 0],
                    [1, 0, 0]],
                   [[2, 0, 0],
                    [0, -2, 0],
                    [0, 0, 0]]])


def paw_correction(density_sii: Array3D,
                   setup: Setup,
                   xc: XCFunctional = None,
                   exclude_core: bool = False) -> Array2D:
    """Corrections from 1-center expansions of spin-density."""
    # Spherical part:
    spin_density_ii = density_sii[0] - density_sii[1]
    D0_jj = expand(spin_density_ii, setup.l_j, l=0)[0]

    phit_jg = np.array(setup.data.phit_jg)
    phi_jg = np.array(setup.data.phi_jg)

    rgd = setup.rgd

    # Spin-density from pseudo density:
    nt0 = phit_jg[:, 0].dot(D0_jj).dot(phit_jg[:, 0]) / (4 * pi)**0.5

    # All-electron contribution diverges as r^-beta and must be integrated
    # over a small region of size rT:
    n0_g = np.einsum('ab, ag, bg -> g', D0_jj, phi_jg, phi_jg) / (4 * pi)**0.5
    if not exclude_core and setup.Nc > 0 and xc is not None:
        n0_g += core_contribution(density_sii, setup, xc)

    beta = 2 * (1 - (1 - (setup.Z * alpha)**2)**0.5)
    rT = setup.Z * alpha**2
    n0 = integrate(n0_g, rgd, rT, beta)

    W1 = (n0 - nt0) * 2 / 3  # isotropic result

    # Now the anisotropic part from the l=2 part of the density:
    D2_mjj = expand(spin_density_ii, setup.l_j, 2)
    dn2_mg = np.einsum('mab, ag, bg -> mg', D2_mjj, phi_jg, phi_jg)
    dn2_mg -= np.einsum('mab, ag, bg -> mg', D2_mjj, phit_jg, phit_jg)
    A_m = dn2_mg[:, 1:].dot(rgd.dr_g[1:] / rgd.r_g[1:]) / 5
    A_m *= Y2_m
    # W_vv: Array2D = Y2_mvv.T.dot(A_m)  # type: ignore
    W_vv = Y2_mvv.T @ A_m
    W = np.trace(W_vv) / 3
    for v in range(3):
        W_vv[v, v] -= W
        W_vv[v, v] += W1

    return W_vv


def expand(D_ii: Array2D,
           l_j: List[int],
           l: int) -> Array3D:
    """Get expansion coefficients."""
    G_LLm = gaunt(lmax=2)[:, :, l**2:(l + 1)**2]
    D_mjj = np.empty((2 * l + 1, len(l_j), len(l_j)))
    i1a = 0
    for j1, l1 in enumerate(l_j):
        i1b = i1a + 2 * l1 + 1
        i2a = 0
        for j2, l2 in enumerate(l_j):
            i2b = i2a + 2 * l2 + 1
            D_mjj[:, j1, j2] = np.einsum('ab, abm -> m',
                                         D_ii[i1a:i1b, i2a:i2b],
                                         G_LLm[l1**2:(l1 + 1)**2,
                                               l2**2:(l2 + 1)**2])
            i2a = i2b
        i1a = i1b
    return D_mjj


def delta(r: Array1D, rT: float) -> Array1D:
    """Extended delta function."""
    return 2 / rT / (1 + 2 * r / rT)**2


def integrate(n0_g: Array1D,
              rgd: RadialGridDescriptor,
              rT: float,
              beta: float) -> float:
    """Take care of r^-beta divergence."""
    r_g = rgd.r_g
    a_i = np.polyfit(r_g[1:5], n0_g[1:5] * r_g[1:5]**beta, 3)
    r4 = r_g[4]
    dr = rT / 50
    n = max(int(r4 / dr / 2) * 2 + 1, 3)
    r_j = np.linspace(0, r4, n)

    b_i = np.arange(3, -1, -1) + 1 - beta
    d0 = 2 / rT  # delta(0, rT)
    d1 = -8 / rT**2
    n0 = a_i.dot(d0 * r4**b_i / b_i + d1 * r4**(b_i + 1) / (b_i + 1))

    d_j = delta(r_j, rT) - (d0 + d1 * r_j)

    head_j = d_j * np.polyval(a_i, r_j)
    head_j[1:] *= r_j[1:]**-beta
    n0 += simpson(head_j, r_j)

    tail_g = n0_g[4:] * delta(r_g[4:], rT)
    if SCIPY_VERSION >= [1, 11]:
        n0 += simpson(tail_g, r_g[4:])
    else:
        n0 += simpson(tail_g, r_g[4:], even='first')

    return n0


def core_contribution(density_sii: Array3D,
                      setup: Setup,
                      xc: XCFunctional) -> Array1D:
    """Calculate spin-density from "frozen" core."""
    # Spherical part:
    D_sjj = [expand(density_ii, setup.l_j, 0)[0]
             for density_ii in density_sii]
    phi_jg = np.array(setup.data.phi_jg)
    rgd = setup.rgd

    # Densities with frozen core:
    n_sg = np.einsum('ag, sab, bg -> sg',
                     phi_jg, D_sjj, phi_jg) / (4 * pi)**0.5
    n_sg += setup.data.nc_g * (0.5 / (4 * pi)**0.5)

    # Potential:
    v_sg = np.zeros_like(n_sg)
    xc.calculate_spherical(rgd, n_sg, v_sg)
    vr_sg = v_sg * rgd.r_g
    vr_sg -= setup.Z
    vr_sg += rgd.poisson(n_sg.sum(axis=0))

    # Find first bound s-state included in PAW potential:
    for n, l in zip(setup.n_j, setup.l_j):
        if l == 0:
            assert n > 0
            n0 = n
            break
    else:
        assert False, (setup.n_j, setup.l_j)

    # Initial guesses for core s-states:
    eigs = [e for n, l, f, e in configurations[setup.symbol][1]
            if n < n0 and l == 0]

    # Solve spherical scalar-relativistic Schrödinger equation:
    core_spin_density_g = rgd.zeros()
    sign = 1.0
    for vr_g in vr_sg:
        channel = Channel(l=0, f_n=[1] * len(eigs))
        channel.e_n = eigs
        channel.phi_ng = rgd.empty(len(eigs))
        channel.solve2(vr_g, rgd=rgd, scalar_relativistic=True, Z=setup.Z)
        assert channel.solve2ok
        core_spin_density_g += sign * channel.calculate_density()
        sign = -1.0

    return core_spin_density_g


# From https://en.wikipedia.org/wiki/Gyromagnetic_ratio
# Units: MHz/T
gyromagnetic_ratios = {'H': (1, 42.577478518),
                       'He': (3, -32.434),
                       'Li': (7, 16.546),
                       'C': (13, 10.7084),
                       'N': (14, 3.077),
                       'O': (17, -5.772),
                       'F': (19, 40.052),
                       'Na': (23, 11.262),
                       'Al': (27, 11.103),
                       'Si': (29, -8.465),
                       'P': (31, 17.235),
                       'Fe': (57, 1.382),
                       'Cu': (63, 11.319),
                       'Zn': (67, 2.669),
                       'Xe': (129, -11.777)}


def main(argv: List[str] = None) -> None:
    """Command-line interface."""
    parser = argparse.ArgumentParser(
        prog='python3 -m gpaw.hyperfine',
        description='Calculate hyperfine parameters.')
    add = parser.add_argument
    add('file', metavar='input-file',
        help='GPW-file (with or without wave functions).')
    add('-g', '--g-factors',
        help='G-factors.  Example: "-g H:5.6,O:-0.76".')
    add('-u', '--units', default='ueV', choices=['ueV', 'MHz'],
        help='Units.  Must be "ueV" (micro-eV, default) or "MHz".')
    add('-x', '--exclude-core', action='store_true')
    add('-d', '--diagonalize', action='store_true',
        help='Show eigenvalues of tensor.')

    args = parser.parse_intermixed_args(argv)

    calc = GPAW(args.file)
    atoms = calc.get_atoms()

    symbols = atoms.symbols
    magmoms = atoms.get_magnetic_moments()
    total_magmom = atoms.get_magnetic_moment()

    g_factors = {symbol: ratio * 1e6 * 4 * pi * units._mp / units._e
                 for symbol, (n, ratio) in gyromagnetic_ratios.items()}

    if args.g_factors:
        for symbol, value in (part.split(':')
                              for part in args.g_factors.split(',')):
            g_factors[symbol] = float(value)

    if args.units == 'ueV':
        scale = 1e6
        unit = 'μeV'
    else:
        scale = units._e / units._hplanck * 1e-6
        unit = 'MHz'

    A_avv = hyperfine_parameters(calc, args.exclude_core)

    print('Hyperfine coupling paramters '
          f'in {unit}:\n')

    if args.diagonalize:
        columns = ['1.', '2.', '3.']
    else:
        columns = ['xx', 'yy', 'zz', 'yz', 'xz', 'xy']
    print('  atom  magmom      ', '       '.join(columns))

    used = {}
    for a, A_vv in enumerate(A_avv):
        symbol = symbols[a]
        magmom = magmoms[a]
        g_factor = g_factors.get(symbol, 1.0)
        used[symbol] = g_factor
        A_vv *= g_factor * scale
        if args.diagonalize:
            numbers = np.linalg.eigvalsh(A_vv).tolist()
        else:
            numbers = [A_vv[0, 0], A_vv[1, 1], A_vv[2, 2],
                       A_vv[1, 2], A_vv[0, 2], A_vv[0, 1]]
        print(f'{a:3} {symbol:>2}  {magmom:6.3f}',
              ''.join(f'{x:9.2f}' for x in numbers))

    print('\nCore correction',
          'NOT included!' if args.exclude_core else 'included')
    print(f'Total magnetic moment: {total_magmom:.3f}')

    print('\nG-factors used:')
    for symbol, g in used.items():
        print(f'{symbol:2} {g:10.3f}')


if __name__ == '__main__':
    main()



import pytest
import numpy as np
from gpaw.grid_descriptor import GridDescriptor
from gpaw.transformers import Transformer
from gpaw.mpi import world
from gpaw.gpu import cupy as cp


@pytest.mark.gpu
@pytest.mark.parametrize('pbc', [True, False])
def test_fd_transformers(gpu, pbc):
    if world.size > 4:
        # Grid is so small that domain decomposition cannot exceed 4 domains
        assert world.size % 4 == 0
        group, other = divmod(world.rank, 4)
        ranks = np.arange(4 * group, 4 * (group + 1))
        domain_comm = world.new_communicator(ranks)
    else:
        domain_comm = world

    lat = 8.0
    gd = GridDescriptor((32, 32, 32), (lat, lat, lat),
                        pbc_c=pbc, comm=domain_comm)

    if pbc:
        dtype = complex
        phase = np.ones((3, 2), complex)
    else:
        dtype = float
        phase = None

    # Use Gaussian as input
    x, y, z = gd.get_grid_point_coordinates()
    sigma = 1.5
    mu = lat / 2.0

    a = gd.zeros(dtype=dtype)
    a[:] = np.exp(-((x - mu)**2 + (y - mu)**2 + (z - mu)**2) / (2.0 * sigma))

    a_gpu = cp.asarray(a)

    # Transformers
    coarsegd = gd.coarsen()
    a_coarse = coarsegd.zeros(dtype=dtype)
    a_coarse_gpu = cp.zeros_like(a_coarse)

    # Restrict
    Transformer(gd, coarsegd, 1, dtype=dtype).apply(a, a_coarse, phases=phase)
    Transformer(gd, coarsegd, 1, dtype=dtype, xp=cp).apply(
        a_gpu, a_coarse_gpu, phases=phase)
    a_coarse_ref = a_coarse_gpu.get()
    assert a_coarse == pytest.approx(a_coarse_ref, abs=1e-14)

    # Interpolate
    Transformer(coarsegd, gd, 1, dtype=dtype).apply(a_coarse, a, phases=phase)
    Transformer(coarsegd, gd, 1, dtype=dtype, xp=cp).apply(
        a_coarse_gpu, a_gpu, phases=phase)
    a_ref = a_gpu.get()
    assert a == pytest.approx(a_ref, abs=1e-14)



from typing import Tuple, Dict

import numpy as np

from gpaw.xc import XC
from .coulomb import coulomb_interaction
from .forces import calculate_forces
from .paw import calculate_paw_stuff
from .scf import apply1, apply2
from .symmetry import Symmetry


class HybridXC:
    orbital_dependent = True
    type = 'HYB'

    def __init__(self,
                 xcname: str,
                 fraction: float = None,
                 omega: float = None):
        from . import parse_name
        if xcname in ['EXX', 'PBE0', 'HSE03', 'HSE06', 'B3LYP']:
            if fraction is not None or omega is not None:
                raise ValueError
            self.name = xcname
            xcname, fraction, omega = parse_name(xcname)
        else:
            if fraction is None or omega is None:
                raise ValueError
            self.name = f'{xcname}-{fraction:.3f}-{omega:.3f}'

        self.xc = XC(xcname)
        self.exx_fraction = fraction
        self.omega = omega

        if xcname == 'null':
            self.description = ''
        else:
            self.description = f'{xcname} + '
        self.description += f'{fraction} * EXX(omega = {omega} bohr^-1)'

        self.vlda_sR = None
        self.v_sknG: Dict[Tuple[int, int], np.ndarray] = {}

        self.ecc = np.nan
        self.evc = np.nan
        self.evv = np.nan
        self.ekin = np.nan

        self.sym = None
        self.coulomb = None

    def get_setup_name(self):
        return 'PBE'

    def initialize(self, dens, ham, wfs):
        self.dens = dens
        self.wfs = wfs
        self.ecc = sum(setup.ExxC for setup in wfs.setups) * self.exx_fraction
        assert wfs.world.size == wfs.gd.comm.size

    def get_description(self):
        return self.description

    def set_positions(self, spos_ac):
        self.spos_ac = spos_ac

    def calculate(self, gd, nt_sr, vt_sr):
        energy = self.ecc + self.evv + self.evc
        energy += self.xc.calculate(gd, nt_sr, vt_sr)
        return energy

    def calculate_paw_correction(self, setup, D_sp, dH_sp=None, a=None):
        return self.xc.calculate_paw_correction(setup, D_sp, dH_sp, a=a)

    def get_kinetic_energy_correction(self):
        return self.ekin

    def apply_orbital_dependent_hamiltonian(self, kpt, psit_xG,
                                            Htpsit_xG=None, dH_asp=None):
        wfs = self.wfs
        if self.coulomb is None:
            self.coulomb = coulomb_interaction(self.omega, wfs.gd, wfs.kd)
            self.description += '\n' + self.coulomb.get_description()
            self.sym = Symmetry(wfs.kd)

        paw_s = calculate_paw_stuff(wfs, self.dens)  # ???????

        if kpt.f_n is None:
            # Just use LDA_X for first step:
            if self.vlda_sR is None:
                # First time:
                self.vlda_sR = self.calculate_lda_potential()
            pd = kpt.psit.pd
            for psit_G, Htpsit_G in zip(psit_xG, Htpsit_xG):
                Htpsit_G += pd.fft(self.vlda_sR[kpt.s] *
                                   pd.ifft(psit_G, kpt.k), kpt.q)
        else:
            self.vlda_sR = None
            if kpt.psit.array.base is psit_xG.base:
                if (kpt.s, kpt.k) not in self.v_sknG:
                    assert not any(s == kpt.s for s, k in self.v_sknG)
                    evc, evv, ekin, v_knG = apply1(
                        kpt, Htpsit_xG,
                        wfs,
                        self.coulomb, self.sym,
                        paw_s[kpt.s])
                    if kpt.s == 0:
                        self.evc = 0.0
                        self.evv = 0.0
                        self.ekin = 0.0
                    scale = 2 / wfs.nspins * self.exx_fraction
                    self.evc += evc * scale
                    self.evv += evv * scale
                    self.ekin += ekin * scale
                    self.v_sknG = {(kpt.s, k): v_nG
                                   for k, v_nG in v_knG.items()}
                v_nG = self.v_sknG.pop((kpt.s, kpt.k))
            else:
                v_nG = apply2(kpt, psit_xG, Htpsit_xG, wfs,
                              self.coulomb, self.sym,
                              paw_s[kpt.s])
            Htpsit_xG += v_nG * self.exx_fraction

    def calculate_lda_potential(self):
        from gpaw.xc import XC
        lda = XC('LDA_X')
        nt_sr = self.dens.nt_sg
        vt_sr = np.zeros_like(nt_sr)
        vlda_sR = self.dens.gd.zeros(self.wfs.nspins)
        lda.calculate(self.dens.finegd, nt_sr, vt_sr)
        for vt_R, vt_r in zip(vlda_sR, vt_sr):
            vt_R[:], _ = self.dens.pd3.restrict(vt_r, self.dens.pd2)
        return vlda_sR * self.exx_fraction

    def summary(self, log):
        log(self.description)

    def add_forces(self, F_av):
        paw_s = calculate_paw_stuff(self.wfs, self.dens)
        F_av += calculate_forces(self.wfs,
                                 self.coulomb,
                                 self.sym,
                                 paw_s) * self.exx_fraction

    def correct_hamiltonian_matrix(self, kpt, H_nn):
        return

    def rotate(self, kpt, U_nn):
        pass  # 1 / 0

    def add_correction(self, kpt, psit_xG, Htpsit_xG, P_axi, c_axi, n_x,
                       calculate_change=False):
        pass  # 1 / 0

    def read(self, reader):
        pass

    def write(self, writer):
        pass

    def set_grid_descriptor(self, gd):
        pass



import numpy as np
from _gpaw import get_num_threads
from ase import Atoms
from ase.data import chemical_symbols
from ase.geometry import cell_to_cellpar
from ase.units import Bohr


def print_cell(gd, pbc_c, log):

    log("""Unit cell:
           periodic     x           y           z      points  spacing""")
    h_c = gd.get_grid_spacings()
    for c in range(3):
        log('  %d. axis:    %s  %10.6f  %10.6f  %10.6f   %3d   %8.4f'
            % ((c + 1, ['no ', 'yes'][int(pbc_c[c])]) +
               tuple(Bohr * gd.cell_cv[c]) +
               (gd.N_c[c], Bohr * h_c[c])))

    par = cell_to_cellpar(gd.cell_cv * Bohr)
    log('\n  Lengths: {:10.6f} {:10.6f} {:10.6f}'.format(*par[:3]))
    log('  Angles:  {:10.6f} {:10.6f} {:10.6f}\n'.format(*par[3:]))

    h_eff = gd.dv**(1.0 / 3.0) * Bohr
    log('Effective grid spacing dv^(1/3) = {:.4f}'.format(h_eff))
    log()


def print_positions(atoms, log, magmom_av):
    log(plot(atoms))
    log('\nPositions:')
    symbols = atoms.get_chemical_symbols()
    for a, pos_v in enumerate(atoms.get_positions()):
        symbol = symbols[a]
        log('{0:>4} {1:3} {2[0]:11.6f} {2[1]:11.6f} {2[2]:11.6f}'
            '    ({3[0]:7.4f}, {3[1]:7.4f}, {3[2]:7.4f})'
            .format(a, symbol, pos_v, magmom_av[a]))
    log()


def print_parallelization_details(wfs, ham, log):
    log('Total number of cores used:', wfs.world.size)
    if wfs.kd.comm.size > 1:
        log('Parallelization over k-points:', wfs.kd.comm.size)

    # Domain decomposition settings:
    coarsesize = tuple(wfs.gd.parsize_c)
    finesize = tuple(ham.finegd.parsize_c)

    try:  # Only planewave density
        xc_gd = ham.xc_gd
    except AttributeError:
        xc_gd = ham.finegd
    xcsize = tuple(xc_gd.parsize_c)

    if any(np.prod(size) != 1 for size in [coarsesize, finesize, xcsize]):
        title = 'Domain decomposition:'
        template = '%d x %d x %d'
        log(title, template % coarsesize)
        if coarsesize != finesize:
            log(' ' * len(title), template % finesize, '(fine grid)')
        if xcsize != finesize:
            log(' ' * len(title), template % xcsize, '(xc only)')

    if wfs.bd.comm.size > 1:  # band parallelization
        log('Parallelization over states: %d' % wfs.bd.comm.size)

    if get_num_threads() > 1:  # OpenMP threading
        log('OpenMP threads: {}'.format(get_num_threads()))
    log()


def plot(atoms: Atoms) -> str:
    """Ascii-art plot of the atoms."""

    #   y
    #   |
    #   .-- x
    #  /
    # z

    if atoms.cell.handedness != 1:
        # See example we can't handle in test_ascii_art()
        return ''

    cell_cv = atoms.get_cell()
    if atoms.cell.orthorhombic:
        plot_box = True
    else:
        atoms = atoms.copy()
        atoms.cell = [1, 1, 1]
        atoms.center(vacuum=2.0)
        cell_cv = atoms.get_cell()
        plot_box = False

    cell = np.diagonal(cell_cv) / Bohr
    positions = atoms.get_positions() / Bohr
    numbers = atoms.get_atomic_numbers()

    s = 1.3
    nx, ny, nz = nxyz = (s * cell * (1.0, 0.25, 0.5) + 0.5).astype(int)
    sx, sy, sz = nxyz / cell
    grid = Grid(nx + ny + 4, nz + ny + 1)
    positions = (positions % cell + cell) % cell
    ij = np.dot(positions, [(sx, 0), (sy, sy), (0, sz)])
    ij = np.around(ij).astype(int)
    for a, Z in enumerate(numbers):
        symbol = chemical_symbols[Z]
        i, j = ij[a]
        depth = positions[a, 1]
        for n, c in enumerate(symbol):
            grid.put(c, i + n + 1, j, depth)
    if plot_box:
        k = 0
        for i, j in [(1, 0), (1 + nx, 0)]:
            grid.put('*', i, j)
            grid.put('.', i + ny, j + ny)
            if k == 0:
                grid.put('*', i, j + nz)
            grid.put('.', i + ny, j + nz + ny)
            for y in range(1, ny):
                grid.put('/', i + y, j + y, y / sy)
                if k == 0:
                    grid.put('/', i + y, j + y + nz, y / sy)
            for z in range(1, nz):
                if k == 0:
                    grid.put('|', i, j + z)
                grid.put('|', i + ny, j + z + ny)
            k = 1
        for i, j in [(1, 0), (1, nz)]:
            for x in range(1, nx):
                if k == 1:
                    grid.put('-', i + x, j)
                grid.put('-', i + x + ny, j + ny)
            k = 0
    return '\n'.join([''.join([chr(x) for x in line])
                      for line in np.transpose(grid.grid)[::-1]])


class Grid:
    def __init__(self, i, j):
        self.grid = np.zeros((i, j), np.int8)
        self.grid[:] = ord(' ')
        self.depth = np.zeros((i, j))
        self.depth[:] = 1e10

    def put(self, c, i, j, depth=1e9):
        if depth < self.depth[i, j]:
            self.grid[i, j] = ord(c)
            self.depth[i, j] = depth



"""Extensions to the ase Atoms class

"""
import numpy as np

from ase import Atoms
from ase.io import read
from ase.build.connected import connected_indices


class Cluster(Atoms):
    """A class for cluster structures
    to enable simplified manipulation"""

    def __init__(self, *args, **kwargs):

        self.data = {}

        if len(args) > 0:
            filename = args[0]
            if isinstance(filename, str):
                self.read(filename, kwargs.get('filetype'))
                return
        else:
            Atoms.__init__(self, [])

        if kwargs.get('filename') is not None:
            filename = kwargs.pop('filename')
            Atoms.__init__(self, *args, **kwargs)
            self.read(filename, kwargs.get('filetype'))
        else:
            Atoms.__init__(self, *args, **kwargs)

    def extreme_positions(self):
        """get the extreme positions of the structure"""
        pos = self.get_positions()
        return np.array([np.minimum.reduce(pos), np.maximum.reduce(pos)])

    def find_connected(self, index, dmax=None, scale=1.5):
        """Find atoms connected to self[index] and return them."""
        return self[connected_indices(self, index, dmax, scale)]

    def minimal_box(self, border=0, h=None, multiple=4):
        """The box needed to fit the structure in.

        The structure is moved to fit into the box [(0,x),(0,y),(0,z)]
        with x,y,z > 0 (fitting the ASE constriction).
        The border argument can be used to add a border of empty space
        around the structure.

        If h is set, the box is extended to ensure that box/h is
        a multiple of 'multiple'.
        This ensures that GPAW uses the desired h.

        The shift applied to the structure is returned.
         """

        if len(self) == 0:
            return None

        extr = self.extreme_positions()

        # add borders
        if isinstance(border, list):
            b = border
        else:
            b = [border, border, border]
        for c in range(3):
            extr[0][c] -= b[c]
            extr[1][c] += b[c] - extr[0][c]  # shifted already

        # check for multiple of 4
        if h is not None:
            if not hasattr(h, '__len__'):
                h = np.array([h, h, h])
            for c in range(3):
                # apply the same as in paw.py
                L = extr[1][c]  # shifted already
                N = np.ceil(L / h[c] / multiple) * multiple
                # correct L
                dL = N * h[c] - L
                # move accordingly
                extr[1][c] += dL  # shifted already
                extr[0][c] -= dL / 2.

        # move lower corner to (0, 0, 0)
        shift = tuple(-1. * np.array(extr[0]))
        self.translate(shift)
        self.set_cell(tuple(extr[1]))

        return shift

    def read(self, filename, format=None):
        """Read the structure from some file. The type can be given
        or it will be guessed from the filename."""

        self.__init__(read(filename, format=format))
        return len(self)
